# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""Resolver output display operation.
"""

__all__ = (
    "Display",
    "format_unmatched_atom",
)

from portage import os
from portage.dbapi.dep_expand import dep_expand
from portage.dep import Atom, cpvequal, _repo_separator, _slot_separator
from portage.eapi import _get_eapi_attrs
from portage.exception import InvalidDependString, SignatureException
from portage.localization import localized_size
from portage.package.ebuild.config import _get_feature_flags
from portage.package.ebuild._spawn_nofetch import spawn_nofetch
from portage.output import (
    blue,
    colorize,
    create_color_func,
    darkblue,
    darkgreen,
    green,
    nc_len,
    teal,
)

bad = create_color_func("BAD")
from portage._sets.base import InternalPackageSet
from portage.util import writemsg_stdout
from portage.versions import best, cpv_getversion

from _emerge.Blocker import Blocker
from _emerge.create_world_atom import create_world_atom
from _emerge.resolver.output_helpers import (
    _DisplayConfig,
    _tree_display,
    _PackageCounters,
    _create_use_string,
    PkgInfo,
)
from _emerge.show_invalid_depstring_notice import show_invalid_depstring_notice


class Display:
    """Formats and outputs the depgraph supplied it for merge/re-merge, etc.

    __call__()
    @param depgraph: list
    @param favorites: defaults to []
    @param verbosity: integer, defaults to None
    """

    def __init__(self):
        self.print_msg = []
        self.blockers = []
        self.counters = _PackageCounters()
        self.resolver = None
        self.resolved = None
        self.vardb = None
        self.portdb = None
        self.verboseadd = ""
        self.oldlp = None
        self.myfetchlist = None
        self.indent = ""
        self.use_expand = None
        self.use_expand_hidden = None
        self.pkgsettings = None
        self.forced_flags = None
        self.newlp = None
        self.conf = None
        self.blocker_style = None

    def _blockers(self, blocker):
        """Adds colorized strings to
        self.print_msg and self.blockers

        @param blocker: _emerge.Blocker.Blocker instance
        @rtype: bool
        Modifies class globals: self.blocker_style, self.resolved,
                self.print_msg
        """
        if blocker.satisfied:
            self.blocker_style = "PKG_BLOCKER_SATISFIED"
            addl = f"{colorize(self.blocker_style, 'b')}     "
        else:
            self.blocker_style = "PKG_BLOCKER"
            addl = f"{colorize(self.blocker_style, 'B')}     "
        addl += self.empty_space_in_brackets()
        self.resolved = dep_expand(
            str(blocker.atom).lstrip("!"), mydb=self.vardb, settings=self.pkgsettings
        )
        if self.conf.columns and self.conf.quiet:
            addl += " " + colorize(self.blocker_style, str(self.resolved))
        else:
            addl = "[{} {}] {}{}".format(
                colorize(self.blocker_style, "blocks"),
                addl,
                self.indent,
                colorize(self.blocker_style, str(self.resolved)),
            )
        block_parents = self.conf.blocker_parents.parent_nodes(blocker)
        block_parents = {str(pnode.cpv) for pnode in block_parents}
        block_parents = ", ".join(block_parents)
        if blocker.atom.blocker.overlap.forbid:
            blocking_desc = "hard blocking"
        else:
            blocking_desc = "soft blocking"
        if self.resolved != blocker.atom:
            addl += colorize(
                self.blocker_style,
                f' ("{str(blocker.atom).lstrip("!")}" is {blocking_desc} {block_parents})',
            )
        else:
            addl += colorize(
                self.blocker_style, f" (is {blocking_desc} {block_parents})"
            )
        if blocker.satisfied:
            if not self.conf.columns:
                self.print_msg.append(addl)
        else:
            self.blockers.append(addl)

    def include_mask_str(self):
        return self.conf.verbosity > 1

    def gen_mask_str(self, pkg):
        """
        @param pkg: _emerge.Package.Package instance
        """
        hardmasked = pkg.isHardMasked()
        mask_str = " "

        if hardmasked:
            mask_str = colorize("BAD", "#")
        else:
            keyword_mask = pkg.get_keyword_mask()

            if keyword_mask is None:
                pass
            elif keyword_mask == "missing":
                mask_str = colorize("BAD", "*")
            else:
                mask_str = colorize("WARN", "~")

        return mask_str

    def empty_space_in_brackets(self):
        space = ""
        if self.include_mask_str():
            # add column for mask status
            space += " "
        return space

    def map_to_use_expand(self, myvals, forced_flags=False, remove_hidden=True):
        """Map use expand variables

        @param myvals: list
        @param forced_flags: bool
        @param remove_hidden: bool
        @rtype ret dictionary
                or ret dict, forced dict.
        """
        ret = {}
        forced = {}
        for exp in self.use_expand:
            ret[exp] = []
            forced[exp] = set()
            for val in myvals[:]:
                if val.startswith(exp.lower() + "_"):
                    if val in self.forced_flags:
                        forced[exp].add(val[len(exp) + 1 :])
                    ret[exp].append(val[len(exp) + 1 :])
                    myvals.remove(val)
        ret["USE"] = myvals
        forced["USE"] = [val for val in myvals if val in self.forced_flags]
        if remove_hidden:
            for exp in self.use_expand_hidden:
                ret.pop(exp, None)
        if forced_flags:
            return ret, forced
        return ret

    def _display_use(self, pkg, pkg_info):
        """USE flag display

        @param pkg: _emerge.Package.Package instance
        @param pkg_info: PkgInfo instance
        Modifies self.use_expand_hidden, self.use_expand, self.verboseadd,
                self.forced_flags
        """

        self.forced_flags = set()
        self.forced_flags.update(pkg.use.force)
        self.forced_flags.update(pkg.use.mask)

        cur_use = [
            flag for flag in self.conf.pkg_use_enabled(pkg) if flag in pkg.iuse.all
        ]
        cur_iuse = sorted(pkg.iuse.all)

        if pkg_info.previous_pkg is not None:
            previous_pkg = pkg_info.previous_pkg
            old_iuse = sorted(previous_pkg.iuse.all)
            old_use = previous_pkg.use.enabled
            is_new = False
        else:
            old_iuse = []
            old_use = []
            is_new = True

        old_use = [flag for flag in old_use if flag in old_iuse]

        self.use_expand = pkg.use.expand
        self.use_expand_hidden = pkg.use.expand_hidden

        # Prevent USE_EXPAND_HIDDEN flags from being hidden if they
        # are the only thing that triggered reinstallation.
        reinst_flags_map = {}
        reinstall_for_flags = self.conf.reinstall_nodes.get(pkg)
        reinst_expand_map = None
        if reinstall_for_flags:
            reinst_flags_map = self.map_to_use_expand(
                list(reinstall_for_flags), remove_hidden=False
            )
            for k in list(reinst_flags_map):
                if not reinst_flags_map[k]:
                    del reinst_flags_map[k]
            if not reinst_flags_map.get("USE"):
                reinst_expand_map = reinst_flags_map.copy()
                reinst_expand_map.pop("USE", None)
        if reinst_expand_map and not set(reinst_expand_map).difference(
            self.use_expand_hidden
        ):
            self.use_expand_hidden = set(self.use_expand_hidden).difference(
                reinst_expand_map
            )

        cur_iuse_map, iuse_forced = self.map_to_use_expand(cur_iuse, forced_flags=True)
        cur_use_map = self.map_to_use_expand(cur_use)
        old_iuse_map = self.map_to_use_expand(old_iuse)
        old_use_map = self.map_to_use_expand(old_use)

        use_expand = sorted(self.use_expand)
        use_expand.insert(0, "USE")
        feature_flags = _get_feature_flags(_get_eapi_attrs(pkg.eapi))

        for key in use_expand:
            if key in self.use_expand_hidden:
                continue
            self.verboseadd += _create_use_string(
                self.conf,
                key.upper(),
                cur_iuse_map[key],
                iuse_forced[key],
                cur_use_map[key],
                old_iuse_map[key],
                old_use_map[key],
                is_new,
                feature_flags,
                reinst_flags_map.get(key),
            )

    @staticmethod
    def pkgprint(pkg_str, pkg_info):
        """Colorizes a string acording to pkg_info settings

        @param pkg_str: string
        @param pkg_info: dictionary
        @rtype colorized string
        """
        if pkg_info.merge:
            if pkg_info.built:
                if pkg_info.system:
                    return colorize("PKG_BINARY_MERGE_SYSTEM", pkg_str)
                if pkg_info.world:
                    return colorize("PKG_BINARY_MERGE_WORLD", pkg_str)
                return colorize("PKG_BINARY_MERGE", pkg_str)

            if pkg_info.system:
                return colorize("PKG_MERGE_SYSTEM", pkg_str)
            if pkg_info.world:
                return colorize("PKG_MERGE_WORLD", pkg_str)
            return colorize("PKG_MERGE", pkg_str)

        if pkg_info.operation == "uninstall":
            return colorize("PKG_UNINSTALL", pkg_str)

        if pkg_info.system:
            return colorize("PKG_NOMERGE_SYSTEM", pkg_str)
        if pkg_info.world:
            return colorize("PKG_NOMERGE_WORLD", pkg_str)
        return colorize("PKG_NOMERGE", pkg_str)

    def verbose_size(self, pkg, repoadd_set, pkg_info):
        """Determines the size of the downloads required

        @param pkg: _emerge.Package.Package instance
        @param repoadd_set: set of repos to add
        @param pkg_info: dictionary
        Modifies class globals: self.myfetchlist, self.counters.totalsize,
                self.verboseadd, repoadd_set.
        """
        mysize = 0
        if pkg.type_name in ("binary", "ebuild") and pkg_info.merge:
            db = pkg.root_config.trees[
                pkg.root_config.pkg_tree_map[pkg.type_name]
            ].dbapi
            kwargs = {}
            if pkg.type_name == "ebuild":
                kwargs["useflags"] = pkg_info.use
                kwargs["myrepo"] = pkg.repo
            myfilesdict = None
            try:
                myfilesdict = db.getfetchsizes(pkg.cpv, **kwargs)
            except InvalidDependString as e:
                # FIXME: validate SRC_URI earlier
                (depstr,) = db.aux_get(pkg.cpv, ["SRC_URI"], myrepo=pkg.repo)
                show_invalid_depstring_notice(pkg, str(e))
                raise
            except SignatureException:
                # missing/invalid binary package SIZE signature
                pass
            if myfilesdict is None:
                myfilesdict = "[empty/missing/bad digest]"
            else:
                for myfetchfile in myfilesdict:
                    if myfetchfile not in self.myfetchlist:
                        mysize += myfilesdict[myfetchfile]
                        self.myfetchlist.add(myfetchfile)
                if pkg_info.ordered:
                    self.counters.totalsize += mysize
            self.verboseadd += localized_size(mysize)

        if self.quiet_repo_display:
            # overlay verbose
            # assign index for a previous version in the same slot
            if pkg_info.previous_pkg is not None:
                repo_name_prev = pkg_info.previous_pkg.repo
            else:
                repo_name_prev = None

            # now use the data to generate output
            if pkg.installed or pkg_info.previous_pkg is None:
                self.repoadd = self.conf.repo_display.repoStr(pkg_info.repo_path_real)
            else:
                repo_path_prev = None
                if repo_name_prev:
                    repo_path_prev = self.portdb.getRepositoryPath(repo_name_prev)
                if repo_path_prev == pkg_info.repo_path_real:
                    self.repoadd = self.conf.repo_display.repoStr(
                        pkg_info.repo_path_real
                    )
                else:
                    self.repoadd = "{}=>{}".format(
                        self.conf.repo_display.repoStr(repo_path_prev),
                        self.conf.repo_display.repoStr(pkg_info.repo_path_real),
                    )
            if self.repoadd:
                repoadd_set.add(self.repoadd)

    def convert_myoldbest(self, pkg, pkg_info):
        """converts and colorizes a version list to a string

        @param pkg: _emerge.Package.Package instance
        @param pkg_info: dictionary
        @rtype string.
        """
        myoldbest = pkg_info.oldbest_list
        # Convert myoldbest from a list to a string.
        myoldbest_str = ""
        if myoldbest:
            versions = []
            for pos, old_pkg in enumerate(myoldbest):
                key = old_pkg.version
                if key[-3:] == "-r0":
                    key = key[:-3]
                if self.conf.verbosity == 3:
                    if pkg_info.attr_display.new_slot:
                        key += _slot_separator + old_pkg.slot
                        if old_pkg.slot != old_pkg.sub_slot:
                            key += "/" + old_pkg.sub_slot
                    elif any(
                        x.slot + "/" + x.sub_slot != "0/0" for x in myoldbest + [pkg]
                    ):
                        key += _slot_separator + old_pkg.slot
                        if (
                            old_pkg.slot != old_pkg.sub_slot
                            or old_pkg.slot == pkg.slot
                            and old_pkg.sub_slot != pkg.sub_slot
                        ):
                            key += "/" + old_pkg.sub_slot
                    if not self.quiet_repo_display:
                        key += _repo_separator + old_pkg.repo
                versions.append(key)
            myoldbest_str = blue("[" + ", ".join(versions) + "]")
        return myoldbest_str

    def _append_slot(self, pkg_str, pkg, pkg_info):
        """Potentially appends slot and subslot to package string.

        @param pkg_str: string
        @param pkg: _emerge.Package.Package instance
        @param pkg_info: dictionary
        @rtype string
        """
        if pkg_info.attr_display.new_slot:
            pkg_str += _slot_separator + pkg_info.slot
            if pkg_info.slot != pkg_info.sub_slot:
                pkg_str += "/" + pkg_info.sub_slot
        elif any(
            x.slot + "/" + x.sub_slot != "0/0" for x in pkg_info.oldbest_list + [pkg]
        ):
            pkg_str += _slot_separator + pkg_info.slot
            if pkg_info.slot != pkg_info.sub_slot or any(
                x.slot == pkg_info.slot and x.sub_slot != pkg_info.sub_slot
                for x in pkg_info.oldbest_list
            ):
                pkg_str += "/" + pkg_info.sub_slot
        return pkg_str

    def _append_repository(self, pkg_str, pkg, pkg_info):
        """Potentially appends repository to package string.

        @param pkg_str: string
        @param pkg: _emerge.Package.Package instance
        @param pkg_info: dictionary
        @rtype string
        """
        if not self.quiet_repo_display:
            pkg_str += _repo_separator + pkg.repo
        return pkg_str

    def _append_build_id(self, pkg_str, pkg, pkg_info):
        """Potentially appends repository to package string.

        @param pkg_str: string
        @param pkg: _emerge.Package.Package instance
        @param pkg_info: dictionary
        @rtype string
        """
        if pkg.type_name == "binary" and pkg.cpv.build_id is not None:
            pkg_str += f"-{pkg.cpv.build_id}"
        return pkg_str

    def _set_non_root_columns(self, pkg, pkg_info):
        """sets the indent level and formats the output

        @param pkg: _emerge.Package.Package instance
        @param pkg_info: dictionary
        @rtype string
        """
        ver_str = self._append_build_id(pkg_info.ver, pkg, pkg_info)
        if self.conf.verbosity == 3:
            ver_str = self._append_slot(ver_str, pkg, pkg_info)
            ver_str = self._append_repository(ver_str, pkg, pkg_info)
        if self.conf.quiet:
            myprint = (
                str(pkg_info.attr_display)
                + " "
                + self.indent
                + self.pkgprint(pkg_info.cp, pkg_info)
            )
            myprint = myprint + darkblue(" " + ver_str) + " "
            myprint = myprint + pkg_info.oldbest
            myprint = myprint + darkgreen("to " + pkg.root)
            self.verboseadd = None
        else:
            if not pkg_info.merge:
                myprint = "[{}] {}{}".format(
                    self.pkgprint(pkg_info.operation.ljust(13), pkg_info),
                    self.indent,
                    self.pkgprint(pkg.cp, pkg_info),
                )
            else:
                myprint = "[{} {}] {}{}".format(
                    self.pkgprint(pkg.type_name, pkg_info),
                    pkg_info.attr_display,
                    self.indent,
                    self.pkgprint(pkg.cp, pkg_info),
                )
            if (self.newlp - nc_len(myprint)) > 0:
                myprint = myprint + (" " * (self.newlp - nc_len(myprint)))
            myprint = myprint + " " + darkblue("[" + ver_str + "]") + " "
            if (self.oldlp - nc_len(myprint)) > 0:
                myprint = myprint + " " * (self.oldlp - nc_len(myprint))
            myprint = myprint + pkg_info.oldbest
            myprint += darkgreen("to " + pkg.root)
        return myprint

    def _set_root_columns(self, pkg, pkg_info):
        """sets the indent level and formats the output

        @param pkg: _emerge.Package.Package instance
        @param pkg_info: dictionary
        @rtype string
        Modifies self.verboseadd
        """
        ver_str = self._append_build_id(pkg_info.ver, pkg, pkg_info)
        if self.conf.verbosity == 3:
            ver_str = self._append_slot(ver_str, pkg, pkg_info)
            ver_str = self._append_repository(ver_str, pkg, pkg_info)
        if self.conf.quiet:
            myprint = (
                str(pkg_info.attr_display)
                + " "
                + self.indent
                + self.pkgprint(pkg_info.cp, pkg_info)
            )
            myprint = myprint + " " + green(ver_str) + " "
            myprint = myprint + pkg_info.oldbest
            self.verboseadd = None
        else:
            if not pkg_info.merge:
                addl = self.empty_space_in_brackets()
                myprint = "[{}{}] {}{}".format(
                    self.pkgprint(pkg_info.operation.ljust(13), pkg_info),
                    addl,
                    self.indent,
                    self.pkgprint(pkg.cp, pkg_info),
                )
            else:
                myprint = "[{} {}] {}{}".format(
                    self.pkgprint(pkg.type_name, pkg_info),
                    pkg_info.attr_display,
                    self.indent,
                    self.pkgprint(pkg.cp, pkg_info),
                )
            if (self.newlp - nc_len(myprint)) > 0:
                myprint = myprint + (" " * (self.newlp - nc_len(myprint)))
            myprint = myprint + " " + green("[" + ver_str + "]") + " "
            if (self.oldlp - nc_len(myprint)) > 0:
                myprint = myprint + (" " * (self.oldlp - nc_len(myprint)))
            myprint += pkg_info.oldbest
        return myprint

    def _set_no_columns(self, pkg, pkg_info):
        """prints pkg info without column indentation.

        @param pkg: _emerge.Package.Package instance
        @param pkg_info: dictionary
        @rtype the updated addl
        """
        pkg_str = self._append_build_id(pkg.cpv, pkg, pkg_info)
        if self.conf.verbosity == 3:
            pkg_str = self._append_slot(pkg_str, pkg, pkg_info)
            pkg_str = self._append_repository(pkg_str, pkg, pkg_info)
        if not pkg_info.merge:
            addl = self.empty_space_in_brackets()
            myprint = "[{}{}] {}{} {}".format(
                self.pkgprint(pkg_info.operation.ljust(13), pkg_info),
                addl,
                self.indent,
                self.pkgprint(pkg_str, pkg_info),
                pkg_info.oldbest,
            )
        else:
            myprint = "[{} {}] {}{} {}".format(
                self.pkgprint(pkg.type_name, pkg_info),
                pkg_info.attr_display,
                self.indent,
                self.pkgprint(pkg_str, pkg_info),
                pkg_info.oldbest,
            )
        return myprint

    def print_messages(self, show_repos):
        """Performs the actual output printing of the pre-formatted
        messages

        @param show_repos: bool.
        """
        for msg in self.print_msg:
            if isinstance(msg, str):
                writemsg_stdout(f"{msg}\n", noiselevel=-1)
                continue
            myprint, self.verboseadd, repoadd = msg
            if self.verboseadd:
                myprint += " " + self.verboseadd
            if show_repos and repoadd:
                myprint += " " + teal(f"[{repoadd}]")
            writemsg_stdout(f"{myprint}\n", noiselevel=-1)

    def print_blockers(self):
        """Performs the actual output printing of the pre-formatted
        blocker messages
        """
        for pkg in self.blockers:
            writemsg_stdout(f"{pkg}\n", noiselevel=-1)

    def print_verbose(self, show_repos):
        """Prints the verbose output to std_out

        @param show_repos: bool.
        """
        writemsg_stdout(f"\n{self.counters}\n", noiselevel=-1)
        if show_repos:
            writemsg_stdout(f"{self.conf.repo_display}", noiselevel=-1)

    def get_display_list(self, mylist):
        """Determines the display list to process

        @param mylist
        @rtype list
        Modifies self.counters.blocks, self.counters.blocks_satisfied,

        """
        unsatisfied_blockers = []
        ordered_nodes = []
        for pkg in mylist:
            if isinstance(pkg, Blocker):
                self.counters.blocks += 1
                if pkg.satisfied:
                    ordered_nodes.append(pkg)
                    self.counters.blocks_satisfied += 1
                else:
                    unsatisfied_blockers.append(pkg)
            else:
                ordered_nodes.append(pkg)
        if self.conf.tree_display:
            display_list = _tree_display(self.conf, ordered_nodes)
        else:
            display_list = [(pkg, 0, True) for pkg in ordered_nodes]
        for pkg in unsatisfied_blockers:
            display_list.append((pkg, 0, True))
        return display_list

    def set_pkg_info(self, pkg, ordered):
        """Sets various pkg_info dictionary variables

        @param pkg: _emerge.Package.Package instance
        @param ordered: bool
        @rtype pkg_info dictionary
        Modifies self.counters.restrict_fetch,
                self.counters.restrict_fetch_satisfied
        """
        pkg_info = PkgInfo()
        pkg_info.cp = pkg.cp
        pkg_info.ver = self.get_ver_str(pkg)
        pkg_info.slot = pkg.slot
        pkg_info.sub_slot = pkg.sub_slot
        pkg_info.repo_name = pkg.repo
        pkg_info.ordered = ordered
        pkg_info.operation = pkg.operation
        pkg_info.merge = ordered and pkg_info.operation == "merge"
        if not pkg_info.merge and pkg_info.operation == "merge":
            pkg_info.operation = "nomerge"
        pkg_info.built = pkg.type_name != "ebuild"
        pkg_info.ebuild_path = None
        if ordered:
            if pkg_info.merge:
                if pkg.type_name == "binary":
                    self.counters.binary += 1
            elif pkg_info.operation == "uninstall":
                self.counters.uninst += 1
        if pkg.type_name == "ebuild":
            pkg_info.ebuild_path = self.portdb.findname(
                pkg.cpv, myrepo=pkg_info.repo_name
            )
            if pkg_info.ebuild_path is None:
                raise AssertionError(f"ebuild not found for '{pkg.cpv}'")
            pkg_info.repo_path_real = os.path.dirname(
                os.path.dirname(os.path.dirname(pkg_info.ebuild_path))
            )
        else:
            pkg_info.repo_path_real = self.portdb.getRepositoryPath(pkg.repo)
        pkg_info.use = list(self.conf.pkg_use_enabled(pkg))
        if not pkg.built and pkg.operation == "merge" and "fetch" in pkg.restrict:
            if pkg_info.ordered:
                self.counters.restrict_fetch += 1
            pkg_info.attr_display.fetch_restrict = True
            if not self.portdb.getfetchsizes(
                pkg.cpv, useflags=pkg_info.use, myrepo=pkg.repo
            ):
                pkg_info.attr_display.fetch_restrict_satisfied = True
                if pkg_info.ordered:
                    self.counters.restrict_fetch_satisfied += 1
            else:
                if pkg_info.ebuild_path is not None:
                    self.restrict_fetch_list[pkg] = pkg_info

        if self.vardb.cpv_exists(pkg.cpv):
            # Do a cpv match first, in case the SLOT has changed.
            pkg_info.previous_pkg = self.vardb.match_pkgs(Atom("=" + pkg.cpv))[0]
        else:
            cp_slot_matches = self.vardb.match_pkgs(pkg.slot_atom)
            if cp_slot_matches:
                pkg_info.previous_pkg = cp_slot_matches[0]
            else:
                cp_matches = self.vardb.match_pkgs(Atom(pkg.cp))
                if cp_matches:
                    # Use highest installed other-slot package instance.
                    pkg_info.previous_pkg = cp_matches[-1]

        return pkg_info

    def check_system_world(self, pkg):
        """Checks for any occurances of the package in the system or world sets

        @param pkg: _emerge.Package.Package instance
        @rtype system and world booleans
        """
        root_config = self.conf.roots[pkg.root]
        system_set = root_config.sets["system"]
        world_set = self.conf.selected_sets[pkg.root]
        system = False
        world = False
        try:
            system = system_set.findAtomForPackage(
                pkg, modified_use=self.conf.pkg_use_enabled(pkg)
            )
            world = world_set.findAtomForPackage(
                pkg, modified_use=self.conf.pkg_use_enabled(pkg)
            )
            if (
                not (self.conf.oneshot or world)
                and pkg.root == self.conf.target_root
                and self.conf.favorites.findAtomForPackage(
                    pkg, modified_use=self.conf.pkg_use_enabled(pkg)
                )
            ):
                # Maybe it will be added to world now.
                if create_world_atom(pkg, self.conf.favorites, root_config):
                    world = True
        except InvalidDependString:
            # This is reported elsewhere if relevant.
            pass
        return system, world

    @staticmethod
    def get_ver_str(pkg):
        """Obtains the version string
        @param pkg: _emerge.Package.Package instance
        @rtype string
        """
        ver_str = pkg.cpv.version
        if ver_str.endswith("-r0"):
            ver_str = ver_str[:-3]
        return ver_str

    def _get_installed_best(self, pkg, pkg_info):
        """we need to use "--emptrytree" testing here rather than
        "empty" param testing because "empty"
        param is used for -u, where you still *do* want to see when
        something is being upgraded.

        @param pkg: _emerge.Package.Package instance
        @param pkg_info: dictionay
        @rtype addl, myoldbest: list, myinslotlist: list
        Modifies self.counters.reinst, self.counters.new

        """
        myoldbest = []
        myinslotlist = None
        installed_versions = self.vardb.match_pkgs(Atom(pkg.cp))
        if self.vardb.cpv_exists(pkg.cpv):
            pkg_info.attr_display.replace = True
            installed_version = pkg_info.previous_pkg
            if (
                installed_version.slot != pkg.slot
                or installed_version.sub_slot != pkg.sub_slot
                or not self.quiet_repo_display
                and installed_version.repo != pkg.repo
            ):
                myoldbest = [installed_version]
            if pkg_info.ordered:
                if pkg_info.merge:
                    self.counters.reinst += 1
        # filter out old-style virtual matches
        elif installed_versions and installed_versions[0].cp == pkg.cp:
            myinslotlist = self.vardb.match_pkgs(pkg.slot_atom)
            # If this is the first install of a new-style virtual, we
            # need to filter out old-style virtual matches.
            if myinslotlist and myinslotlist[0].cp != pkg.cp:
                myinslotlist = None
            if myinslotlist:
                myoldbest = myinslotlist[:]
                if not cpvequal(
                    pkg.cpv, best([pkg.cpv] + [x.cpv for x in myinslotlist])
                ):
                    # Downgrade in slot
                    pkg_info.attr_display.new_version = True
                    pkg_info.attr_display.downgrade = True
                    if pkg_info.ordered:
                        self.counters.downgrades += 1
                else:
                    # Update in slot
                    pkg_info.attr_display.new_version = True
                    if pkg_info.ordered:
                        self.counters.upgrades += 1
            else:
                myoldbest = installed_versions
                pkg_info.attr_display.new = True
                pkg_info.attr_display.new_slot = True
                if pkg_info.ordered:
                    self.counters.newslot += 1
        else:
            pkg_info.attr_display.new = True
            if pkg_info.ordered:
                self.counters.new += 1
        return myoldbest, myinslotlist

    def __call__(self, depgraph, mylist, favorites=None, verbosity=None):
        """The main operation to format and display the resolver output.

        @param depgraph: dependency grah
        @param mylist: list of packages being processed
        @param favorites: list, defaults to []
        @param verbosity: verbose level, defaults to None
        Modifies self.conf, self.myfetchlist, self.portdb, self.vardb,
                self.pkgsettings, self.verboseadd, self.oldlp, self.newlp,
                self.print_msg,
        """
        if favorites is None:
            favorites = []
        self.conf = _DisplayConfig(depgraph, mylist, favorites, verbosity)
        mylist = self.get_display_list(self.conf.mylist)
        # files to fetch list - avoids counting a same file twice
        # in size display (verbose mode)
        self.myfetchlist = set()

        self.quiet_repo_display = (
            "--quiet-repo-display" in depgraph._frozen_config.myopts
        )
        if self.quiet_repo_display:
            # Use this set to detect when all the "repoadd" strings are "[0]"
            # and disable the entire repo display in this case.
            repoadd_set = set()

        self.restrict_fetch_list = {}

        for mylist_index in range(len(mylist)):
            pkg, depth, ordered = mylist[mylist_index]
            self.portdb = self.conf.trees[pkg.root]["porttree"].dbapi
            self.vardb = self.conf.trees[pkg.root]["vartree"].dbapi
            self.pkgsettings = self.conf.pkgsettings[pkg.root]
            self.indent = " " * depth

            if isinstance(pkg, Blocker):
                self._blockers(pkg)
            else:
                pkg_info = self.set_pkg_info(pkg, ordered)
                pkg_info.oldbest_list, myinslotlist = self._get_installed_best(
                    pkg, pkg_info
                )
                if ordered and pkg_info.merge and not pkg_info.attr_display.new:
                    for arg, atom in depgraph._iter_atoms_for_pkg(pkg):
                        if arg.force_reinstall:
                            pkg_info.attr_display.force_reinstall = True
                            break

                self.verboseadd = ""
                if self.quiet_repo_display:
                    self.repoadd = None
                self._display_use(pkg, pkg_info)
                if self.conf.verbosity == 3:
                    if self.quiet_repo_display:
                        self.verbose_size(pkg, repoadd_set, pkg_info)
                    else:
                        self.verbose_size(pkg, None, pkg_info)

                self.oldlp = self.conf.columnwidth - 30
                self.newlp = self.oldlp - 30
                pkg_info.oldbest = self.convert_myoldbest(pkg, pkg_info)
                pkg_info.system, pkg_info.world = self.check_system_world(pkg)
                if "interactive" in pkg.properties and pkg.operation == "merge":
                    pkg_info.attr_display.interactive = True
                    if ordered:
                        self.counters.interactive += 1

                if self.include_mask_str():
                    pkg_info.attr_display.mask = self.gen_mask_str(pkg)

                if pkg.root_config.settings["ROOT"] != "/":
                    if pkg_info.oldbest:
                        pkg_info.oldbest += " "
                    if self.conf.columns:
                        myprint = self._set_non_root_columns(pkg, pkg_info)
                    else:
                        pkg_str = self._append_build_id(pkg.cpv, pkg, pkg_info)
                        if self.conf.verbosity == 3:
                            pkg_str = self._append_slot(pkg_str, pkg, pkg_info)
                            pkg_str = self._append_repository(pkg_str, pkg, pkg_info)
                        if not pkg_info.merge:
                            addl = self.empty_space_in_brackets()
                            myprint = "[{}{}] ".format(
                                self.pkgprint(pkg_info.operation.ljust(13), pkg_info),
                                addl,
                            )
                        else:
                            myprint = "[{} {}] ".format(
                                self.pkgprint(pkg.type_name, pkg_info),
                                pkg_info.attr_display,
                            )
                        myprint += (
                            self.indent
                            + self.pkgprint(pkg_str, pkg_info)
                            + " "
                            + pkg_info.oldbest
                            + darkgreen("to " + pkg.root)
                        )
                else:
                    if self.conf.columns:
                        myprint = self._set_root_columns(pkg, pkg_info)
                    else:
                        myprint = self._set_no_columns(pkg, pkg_info)

                if self.conf.columns and pkg.operation == "uninstall":
                    continue
                if self.quiet_repo_display:
                    self.print_msg.append((myprint, self.verboseadd, self.repoadd))
                else:
                    self.print_msg.append((myprint, self.verboseadd, None))

        show_repos = self.quiet_repo_display and repoadd_set and repoadd_set != {"0"}

        # now finally print out the messages
        self.print_messages(show_repos)
        self.print_blockers()
        if self.conf.verbosity == 3:
            self.print_verbose(show_repos)
        for pkg, pkg_info in self.restrict_fetch_list.items():
            writemsg_stdout(f"\nFetch instructions for {pkg.cpv}:\n", noiselevel=-1)
            spawn_nofetch(
                self.conf.trees[pkg.root]["porttree"].dbapi, pkg_info.ebuild_path
            )

        return os.EX_OK


def format_unmatched_atom(pkg, atom, pkg_use_enabled):
    """
    Returns two strings. The first string contains the
    'atom' with parts of the atom colored, which 'pkg'
    doesn't match. The second string has the same number
    of characters as the first one, but consists of only
    white space or ^. The ^ characters have the same position
    as the colored parts of the first string.
    """
    # Things to check:
    # 	1. Version
    # 	2. cp
    #   3. slot/sub_slot
    # 	4. repository
    # 	5. USE

    if atom.soname:
        return f"{atom}", ""

    highlight = set()

    def perform_coloring():
        atom_str = ""
        marker_str = ""
        for ii, x in enumerate(atom):
            if ii in highlight:
                atom_str += colorize("BAD", x)
                marker_str += "^"
            else:
                atom_str += x
                marker_str += " "
        return atom_str, marker_str

    if atom.cp != pkg.cp:
        # Highlight the cp part only.
        ii = atom.find(atom.cp)
        highlight.update(range(ii, ii + len(atom.cp)))
        return perform_coloring()

    version_atom = atom.without_repo.without_slot.without_use
    version_atom_set = InternalPackageSet(initial_atoms=(version_atom,))
    highlight_version = not bool(
        version_atom_set.findAtomForPackage(pkg, modified_use=pkg_use_enabled(pkg))
    )

    highlight_slot = False
    if (atom.slot and atom.slot != pkg.slot) or (
        atom.sub_slot and atom.sub_slot != pkg.sub_slot
    ):
        highlight_slot = True

    if highlight_version:
        op = atom.operator
        ver = None
        if atom.cp != atom.cpv:
            ver = cpv_getversion(atom.cpv)

        if op == "=*":
            op = "="
            ver += "*"

        if op is not None:
            highlight.update(range(len(op)))

        if ver is not None:
            start = atom.rfind(ver)
            end = start + len(ver)
            highlight.update(range(start, end))

    if highlight_slot:
        slot_str = ":" + atom.slot
        if atom.sub_slot:
            slot_str += "/" + atom.sub_slot
        if atom.slot_operator:
            slot_str += atom.slot_operator
        start = atom.find(slot_str)
        end = start + len(slot_str)
        highlight.update(range(start, end))

    highlight_use = set()
    if atom.use:
        use_atom = f"{atom.cp}[{str(atom.use)}]"
        use_atom_set = InternalPackageSet(initial_atoms=(use_atom,))
        if not use_atom_set.findAtomForPackage(pkg, modified_use=pkg_use_enabled(pkg)):
            missing_iuse = pkg.iuse.get_missing_iuse(atom.unevaluated_atom.use.required)
            if missing_iuse:
                highlight_use = set(missing_iuse)
            else:
                # Use conditionals not met.
                violated_atom = atom.violated_conditionals(
                    pkg_use_enabled(pkg), pkg.iuse.is_valid_flag
                )
                if violated_atom.use is not None:
                    highlight_use = set(
                        violated_atom.use.enabled.union(violated_atom.use.disabled)
                    )

    if highlight_use:
        ii = atom.find("[") + 1
        for token in atom.use.tokens:
            if token.lstrip("-!").rstrip("=?") in highlight_use:
                highlight.update(range(ii, ii + len(token)))
            ii += len(token) + 1

    return perform_coloring()



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2


class DummyTree:
    """
    Most internal code only accesses the "dbapi" attribute of the
    binarytree, portagetree, and vartree classes. DummyTree is useful
    in cases where alternative dbapi implementations (or wrappers that
    modify or extend behavior of existing dbapi implementations) are
    needed, since it allows these implementations to be exposed through
    an interface which is minimally compatible with the *tree classes.
    """

    __slots__ = ("dbapi",)

    def __init__(self, dbapi):
        self.dbapi = dbapi



# Copyright 2014-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotConflictWithBlockerTestCase(TestCase):
    def testBlocker(self):
        ebuilds = {
            "dev-libs/A-1": {"DEPEND": "dev-libs/X"},
            "dev-libs/B-1": {"DEPEND": "<dev-libs/X-2"},
            "dev-libs/C-1": {"DEPEND": "<dev-libs/X-3"},
            "dev-libs/X-1": {"EAPI": "2", "RDEPEND": "!=dev-libs/Y-1"},
            "dev-libs/X-2": {"EAPI": "2", "RDEPEND": "!=dev-libs/Y-2"},
            "dev-libs/X-3": {"EAPI": "2", "RDEPEND": "!=dev-libs/Y-3"},
            "dev-libs/Y-1": {"SLOT": "1"},
            "dev-libs/Y-2": {"SLOT": "2"},
            "dev-libs/Y-3": {"SLOT": "3"},
        }

        installed = {
            "dev-libs/Y-1": {"SLOT": "1"},
            "dev-libs/Y-2": {"SLOT": "2"},
            "dev-libs/Y-3": {"SLOT": "3"},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A", "dev-libs/B", "dev-libs/C"],
                options={"--backtrack": 0},
                all_permutations=True,
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    "dev-libs/X-1",
                    "[uninstall]dev-libs/Y-1",
                    "!=dev-libs/Y-1",
                    ("dev-libs/A-1", "dev-libs/B-1", "dev-libs/C-1"),
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBlockerBuildpkgonly(self):
        ebuilds = {
            "dev-libs/A-1": {"EAPI": "7", "DEPEND": "!!dev-libs/X"},
            "dev-libs/B-1": {"EAPI": "7", "BDEPEND": "!!dev-libs/X"},
            "dev-libs/C-1": {"EAPI": "7", "BDEPEND": "!dev-libs/X"},
            "dev-libs/D-1": {"EAPI": "7", "DEPEND": "!dev-libs/X"},
            "dev-libs/E-1": {"EAPI": "7", "RDEPEND": "!dev-libs/X !!dev-libs/X"},
            "dev-libs/F-1": {"EAPI": "7", "PDEPEND": "!dev-libs/X !!dev-libs/X"},
        }

        installed = {
            "dev-libs/X-1": {},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                success=False,
                options={"--buildpkgonly": True},
                mergelist=["dev-libs/A-1", "!!dev-libs/X"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                success=False,
                options={"--buildpkgonly": True},
                mergelist=["dev-libs/B-1", "!!dev-libs/X"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/C"],
                success=True,
                options={"--buildpkgonly": True},
                mergelist=["dev-libs/C-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/D"],
                success=True,
                options={"--buildpkgonly": True},
                mergelist=["dev-libs/D-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/E"],
                success=True,
                options={"--buildpkgonly": True},
                mergelist=["dev-libs/E-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/F"],
                success=True,
                options={"--buildpkgonly": True},
                mergelist=["dev-libs/F-1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import collections
import fnmatch
import functools
from itertools import chain
import os
import re

from portage.util import shlex_split
from portage.util import (
    normalize_path,
    varexpand,
)


class SonameDepsProcessor:
    """
    Processes NEEDED.ELF.2 entries for one package, in order to generate
    REQUIRES and PROVIDES metadata.

    Any sonames provided by the package will automatically be filtered
    from the generated REQUIRES values.
    """

    def __init__(self, provides_exclude, requires_exclude):
        """
        @param provides_exclude: PROVIDES_EXCLUDE value
        @type provides_exclude: str
        @param requires_exclude: REQUIRES_EXCLUDE value
        @type requires_exclude: str
        """
        self._provides_exclude = self._exclude_pattern(provides_exclude)
        self._requires_exclude = self._exclude_pattern(requires_exclude)
        self._requires_map = collections.defaultdict(
            functools.partial(collections.defaultdict, set)
        )
        self._provides_map = {}
        self._provides_unfiltered = {}
        self._basename_map = {}
        self._provides = None
        self._requires = None
        self._intersected = False

    @staticmethod
    def _exclude_pattern(s):
        # shlex_split enables quoted whitespace inside patterns
        if s:
            pat = re.compile(
                "|".join(fnmatch.translate(x.lstrip(os.sep)) for x in shlex_split(s))
            )
        else:
            pat = None
        return pat

    def add(self, entry):
        """
        Add one NEEDED.ELF.2 entry, for inclusion in the generated
        REQUIRES and PROVIDES values.

        @param entry: NEEDED.ELF.2 entry
        @type entry: NeededEntry
        """

        multilib_cat = entry.multilib_category
        if multilib_cat is None:
            # This usage is invalid. The caller must ensure that
            # the multilib category data is supplied here.
            raise AssertionError(f"Missing multilib category data: {entry.filename}")

        self._basename_map.setdefault(os.path.basename(entry.filename), []).append(
            entry
        )

        if entry.needed and (
            self._requires_exclude is None
            or self._requires_exclude.match(entry.filename.lstrip(os.sep)) is None
        ):
            runpaths = frozenset()
            if entry.runpaths is not None:
                expand = {"ORIGIN": os.path.dirname(entry.filename)}
                runpaths = frozenset(
                    normalize_path(
                        varexpand(
                            x,
                            expand,
                            error_leader=lambda: f"{entry.filename}: DT_RUNPATH: ",
                        )
                    )
                    for x in entry.runpaths
                )
            for x in entry.needed:
                if (
                    self._requires_exclude is None
                    or self._requires_exclude.match(x) is None
                ):
                    self._requires_map[multilib_cat][x].add(runpaths)

        if entry.soname:
            self._provides_unfiltered.setdefault(multilib_cat, set()).add(entry.soname)

        if entry.soname and (
            self._provides_exclude is None
            or (
                self._provides_exclude.match(entry.filename.lstrip(os.sep)) is None
                and self._provides_exclude.match(entry.soname) is None
            )
        ):
            self._provides_map.setdefault(multilib_cat, set()).add(entry.soname)

    def _intersect(self):
        requires_map = self._requires_map
        provides_map = self._provides_map
        provides_unfiltered = self._provides_unfiltered

        for multilib_cat in set(chain(requires_map, provides_map)):
            provides_map.setdefault(multilib_cat, set())
            provides_unfiltered.setdefault(multilib_cat, set())
            for soname, consumers in list(requires_map[multilib_cat].items()):
                if soname in provides_unfiltered[multilib_cat]:
                    del requires_map[multilib_cat][soname]
                elif soname in self._basename_map:
                    # Handle internal libraries that lack an soname, which
                    # are resolved via DT_RUNPATH, see ebtables for example
                    # (bug 646190).
                    for entry in self._basename_map[soname]:
                        if entry.multilib_category != multilib_cat:
                            continue
                        dirname = os.path.dirname(entry.filename)
                        for runpaths in list(consumers):
                            if dirname in runpaths:
                                consumers.remove(runpaths)
                        if not consumers:
                            del requires_map[multilib_cat][soname]
                            break

        provides_data = []
        for multilib_cat in sorted(provides_map):
            if provides_map[multilib_cat]:
                provides_data.append(multilib_cat + ":")
                provides_data.extend(sorted(provides_map[multilib_cat]))

        if provides_data:
            self._provides = " ".join(provides_data) + "\n"

        requires_data = []
        for multilib_cat in sorted(requires_map):
            if requires_map[multilib_cat]:
                requires_data.append(multilib_cat + ":")
                requires_data.extend(sorted(requires_map[multilib_cat]))

        if requires_data:
            self._requires = " ".join(requires_data) + "\n"

        self._intersected = True

    @property
    def provides(self):
        """
        @rtype: str
        @return: PROVIDES value generated from NEEDED.ELF.2 entries
        """
        if not self._intersected:
            self._intersect()
        return self._provides

    @property
    def requires(self):
        """
        @rtype: str
        @return: REQUIRES value generated from NEEDED.ELF.2 entries
        """
        if not self._intersected:
            self._intersect()
        return self._requires



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["dep_check", "dep_eval", "dep_wordreduce", "dep_zapdeps"]

import collections
import itertools
import logging
import operator

import portage
from portage.dep import Atom, match_from_list, use_reduce
from portage.dep._dnf import (
    dnf_convert as _dnf_convert,
    contains_disjunction as _contains_disjunction,
)
from portage.exception import InvalidDependString, ParseError
from portage.localization import _
from portage.util import writemsg, writemsg_level
from portage.util.digraph import digraph
from portage.util.SlotObject import SlotObject
from portage.versions import vercmp


def _expand_new_virtuals(
    mysplit,
    edebug,
    mydbapi,
    mysettings,
    myroot="/",
    trees=None,
    use_mask=None,
    use_force=None,
    **kwargs,
):
    """
    In order to solve bug #141118, recursively expand new-style virtuals so
    as to collapse one or more levels of indirection, generating an expanded
    search space. In dep_zapdeps, new-style virtuals will be assigned
    zero cost regardless of whether or not they are currently installed. Virtual
    blockers are supported but only when the virtual expands to a single
    atom because it wouldn't necessarily make sense to block all the components
    of a compound virtual.  When more than one new-style virtual is matched,
    the matches are sorted from highest to lowest versions and the atom is
    expanded to || ( highest match ... lowest match ).

    The result is normalized in the same way as use_reduce, having a top-level
    conjunction, and no redundant nested lists.
    """
    newsplit = []
    mytrees = trees[myroot]
    portdb = mytrees["porttree"].dbapi
    pkg_use_enabled = mytrees.get("pkg_use_enabled")
    # Atoms are stored in the graph as (atom, id(atom)) tuples
    # since each atom is considered to be a unique entity. For
    # example, atoms that appear identical may behave differently
    # in USE matching, depending on their unevaluated form. Also,
    # specially generated virtual atoms may appear identical while
    # having different _orig_atom attributes.
    atom_graph = mytrees.get("atom_graph")
    parent = mytrees.get("parent")
    virt_parent = mytrees.get("virt_parent")
    graph_parent = None
    if parent is not None:
        if virt_parent is not None:
            graph_parent = virt_parent
            parent = virt_parent
        else:
            graph_parent = parent
    repoman = not mysettings.local_config
    if kwargs["use_binaries"]:
        portdb = trees[myroot]["bintree"].dbapi
    pprovideddict = mysettings.pprovideddict
    myuse = kwargs["myuse"]
    is_disjunction = mysplit and mysplit[0] == "||"
    for x in mysplit:
        if x == "||":
            newsplit.append(x)
            continue
        elif isinstance(x, list):
            assert x, f"Normalization error, empty conjunction found in {mysplit}"
            if is_disjunction:
                assert (
                    x[0] != "||"
                ), f"Normalization error, nested disjunction found in {mysplit}"
            else:
                assert (
                    x[0] == "||"
                ), f"Normalization error, nested conjunction found in {mysplit}"
            x_exp = _expand_new_virtuals(
                x,
                edebug,
                mydbapi,
                mysettings,
                myroot=myroot,
                trees=trees,
                use_mask=use_mask,
                use_force=use_force,
                **kwargs,
            )
            if is_disjunction:
                if len(x_exp) == 1:
                    x = x_exp[0]
                    if isinstance(x, list):
                        # Due to normalization, a conjunction must not be
                        # nested directly in another conjunction, so this
                        # must be a disjunction.
                        assert (
                            x and x[0] == "||"
                        ), f"Normalization error, nested conjunction found in {x_exp}"
                        newsplit.extend(x[1:])
                    else:
                        newsplit.append(x)
                else:
                    newsplit.append(x_exp)
            else:
                newsplit.extend(x_exp)
            continue

        if not isinstance(x, Atom):
            raise ParseError(_("invalid token: '%s'") % x)

        if repoman:
            x = x._eval_qa_conditionals(use_mask, use_force)

        mykey = x.cp
        if not mykey.startswith("virtual/"):
            newsplit.append(x)
            if atom_graph is not None:
                atom_graph.add((x, id(x)), graph_parent)
            continue

        if x.blocker:
            # Virtual blockers are no longer expanded here since
            # the un-expanded virtual atom is more useful for
            # maintaining a cache of blocker atoms.
            newsplit.append(x)
            if atom_graph is not None:
                atom_graph.add((x, id(x)), graph_parent)
            continue

        if repoman or not hasattr(portdb, "match_pkgs") or pkg_use_enabled is None:
            if portdb.cp_list(x.cp):
                newsplit.append(x)
            else:
                a = []
                myvartree = mytrees.get("vartree")
                if myvartree is not None:
                    mysettings._populate_treeVirtuals_if_needed(myvartree)
                mychoices = mysettings.getvirtuals().get(mykey, [])
                for y in mychoices:
                    a.append(Atom(x.replace(x.cp, y.cp, 1)))
                if not a:
                    newsplit.append(x)
                elif is_disjunction:
                    newsplit.extend(a)
                elif len(a) == 1:
                    newsplit.append(a[0])
                else:
                    newsplit.append(["||"] + a)
            continue

        pkgs = []
        # Ignore USE deps here, since otherwise we might not
        # get any matches. Choices with correct USE settings
        # will be preferred in dep_zapdeps().
        matches = portdb.match_pkgs(x.without_use)
        # Use descending order to prefer higher versions.
        matches.reverse()
        for pkg in matches:
            # only use new-style matches
            if pkg.cp.startswith("virtual/"):
                pkgs.append(pkg)

        mychoices = []
        if not pkgs and not portdb.cp_list(x.cp):
            myvartree = mytrees.get("vartree")
            if myvartree is not None:
                mysettings._populate_treeVirtuals_if_needed(myvartree)
            mychoices = mysettings.getvirtuals().get(mykey, [])

        if not (pkgs or mychoices):
            # This one couldn't be expanded as a new-style virtual.  Old-style
            # virtuals have already been expanded by dep_virtual, so this one
            # is unavailable and dep_zapdeps will identify it as such.  The
            # atom is not eliminated here since it may still represent a
            # dependency that needs to be satisfied.
            newsplit.append(x)
            if atom_graph is not None:
                atom_graph.add((x, id(x)), graph_parent)
            continue

        a = []
        for pkg in pkgs:
            virt_atom = "=" + pkg.cpv
            if x.unevaluated_atom.use:
                virt_atom += str(x.unevaluated_atom.use)
                virt_atom = Atom(virt_atom)
                if parent is None:
                    if myuse is None:
                        virt_atom = virt_atom.evaluate_conditionals(
                            mysettings.get("PORTAGE_USE", "").split()
                        )
                    else:
                        virt_atom = virt_atom.evaluate_conditionals(myuse)
                else:
                    virt_atom = virt_atom.evaluate_conditionals(pkg_use_enabled(parent))
            else:
                virt_atom = Atom(virt_atom)

            # Allow the depgraph to map this atom back to the
            # original, in order to avoid distortion in places
            # like display or conflict resolution code.
            virt_atom.__dict__["_orig_atom"] = x

            # According to GLEP 37, RDEPEND is the only dependency
            # type that is valid for new-style virtuals. Repoman
            # should enforce this.
            depstring = pkg._metadata["RDEPEND"]
            pkg_kwargs = kwargs.copy()
            pkg_kwargs["myuse"] = pkg_use_enabled(pkg)
            if edebug:
                writemsg_level(
                    _("Virtual Parent:      %s\n") % (pkg,),
                    noiselevel=-1,
                    level=logging.DEBUG,
                )
                writemsg_level(
                    _("Virtual Depstring:   %s\n") % (depstring,),
                    noiselevel=-1,
                    level=logging.DEBUG,
                )

            # Set EAPI used for validation in dep_check() recursion.
            mytrees["virt_parent"] = pkg

            try:
                mycheck = dep_check(
                    depstring,
                    mydbapi,
                    mysettings,
                    myroot=myroot,
                    trees=trees,
                    **pkg_kwargs,
                )
            finally:
                # Restore previous EAPI after recursion.
                if virt_parent is not None:
                    mytrees["virt_parent"] = virt_parent
                else:
                    del mytrees["virt_parent"]

            if not mycheck[0]:
                raise ParseError(f"{pkg}: {mycheck[1]} '{depstring}'")

            # Replace the original atom "x" with "virt_atom" which refers
            # to the specific version of the virtual whose deps we're
            # expanding. The virt_atom._orig_atom attribute is used
            # by depgraph to map virt_atom back to the original atom.
            # We specifically exclude the original atom "x" from the
            # the expanded output here, since otherwise it could trigger
            # incorrect dep_zapdeps behavior (see bug #597752).
            mycheck[1].append(virt_atom)
            a.append(mycheck[1])
            if atom_graph is not None:
                virt_atom_node = (virt_atom, id(virt_atom))
                atom_graph.add(virt_atom_node, graph_parent)
                atom_graph.add(pkg, virt_atom_node)
                atom_graph.add((x, id(x)), graph_parent)

        if not a and mychoices:
            # Check for a virtual package.provided match.
            for y in mychoices:
                new_atom = Atom(x.replace(x.cp, y.cp, 1))
                if match_from_list(new_atom, pprovideddict.get(new_atom.cp, [])):
                    a.append(new_atom)
                    if atom_graph is not None:
                        atom_graph.add((new_atom, id(new_atom)), graph_parent)

        if not a:
            newsplit.append(x)
            if atom_graph is not None:
                atom_graph.add((x, id(x)), graph_parent)
        elif is_disjunction:
            newsplit.extend(a)
        elif len(a) == 1:
            newsplit.extend(a[0])
        else:
            newsplit.append(["||"] + a)

    # For consistency with related functions like use_reduce, always
    # normalize the result to have a top-level conjunction.
    if is_disjunction:
        newsplit = [newsplit]

    return newsplit


def dep_eval(deplist):
    if not deplist:
        return 1
    if deplist[0] == "||":
        # or list; we just need one "1"
        for x in deplist[1:]:
            if isinstance(x, list):
                if dep_eval(x) == 1:
                    return 1
            elif x == 1:
                return 1
        # XXX: unless there's no available atoms in the list
        # in which case we need to assume that everything is
        # okay as some ebuilds are relying on an old bug.
        if len(deplist) == 1:
            return 1
        return 0
    for x in deplist:
        if isinstance(x, list):
            if dep_eval(x) == 0:
                return 0
        elif x == 0 or x == 2:
            return 0
    return 1


class _dep_choice(SlotObject):
    __slots__ = (
        "atoms",
        "slot_map",
        "cp_map",
        "all_available",
        "all_installed_slots",
        "new_slot_count",
        "want_update",
        "all_in_graph",
    )


def dep_zapdeps(
    unreduced, reduced, myroot, use_binaries=0, trees=None, minimize_slots=False
):
    """
    Takes an unreduced and reduced deplist and removes satisfied dependencies.
    Returned deplist contains steps that must be taken to satisfy dependencies.
    """
    if trees is None:
        trees = portage.db
    writemsg(f"ZapDeps -- {use_binaries}\n", 2)
    if not reduced or unreduced == ["||"] or dep_eval(reduced):
        return []

    if unreduced[0] != "||":
        unresolved = []
        for x, satisfied in zip(unreduced, reduced):
            if isinstance(x, list):
                unresolved += dep_zapdeps(
                    x,
                    satisfied,
                    myroot,
                    use_binaries=use_binaries,
                    trees=trees,
                    minimize_slots=minimize_slots,
                )
            elif not satisfied:
                unresolved.append(x)
        return unresolved

    # We're at a ( || atom ... ) type level and need to make a choice
    deps = unreduced[1:]
    satisfieds = reduced[1:]

    # Our preference order is for an the first item that:
    # a) contains all unmasked packages with the same key as installed packages
    # b) contains all unmasked packages
    # c) contains masked installed packages
    # d) is the first item

    preferred_in_graph = []
    preferred_installed = preferred_in_graph
    preferred_any_slot = preferred_in_graph
    preferred_non_installed = []
    unsat_use_in_graph = []
    unsat_use_installed = []
    unsat_use_non_installed = []
    other_installed = []
    other_installed_some = []
    other_installed_any_slot = []
    other = []

    # unsat_use_* must come after preferred_non_installed
    # for correct ordering in cases like || ( foo[a] foo[b] ).
    choice_bins = (
        preferred_in_graph,
        preferred_non_installed,
        unsat_use_in_graph,
        unsat_use_installed,
        unsat_use_non_installed,
        other_installed,
        other_installed_some,
        other_installed_any_slot,
        other,
    )

    # Alias the trees we'll be checking availability against
    parent = trees[myroot].get("parent")
    virt_parent = trees[myroot].get("virt_parent")
    priority = trees[myroot].get("priority")
    graph_db = trees[myroot].get("graph_db")
    graph = trees[myroot].get("graph")
    pkg_use_enabled = trees[myroot].get("pkg_use_enabled")
    graph_interface = trees[myroot].get("graph_interface")
    downgrade_probe = trees[myroot].get("downgrade_probe")
    circular_dependency = trees[myroot].get("circular_dependency")
    vardb = None
    if "vartree" in trees[myroot]:
        vardb = trees[myroot]["vartree"].dbapi
    if use_binaries:
        mydbapi = trees[myroot]["bintree"].dbapi
    else:
        mydbapi = trees[myroot]["porttree"].dbapi

    try:
        mydbapi_match_pkgs = mydbapi.match_pkgs
    except AttributeError:

        def mydbapi_match_pkgs(atom):
            return [mydbapi._pkg_str(cpv, atom.repo) for cpv in mydbapi.match(atom)]

    # Sort the deps into installed, not installed but already
    # in the graph and other, not installed and not in the graph
    # and other, with values of [[required_atom], availability]
    for x, satisfied in zip(deps, satisfieds):
        if isinstance(x, list):
            atoms = dep_zapdeps(
                x,
                satisfied,
                myroot,
                use_binaries=use_binaries,
                trees=trees,
                minimize_slots=minimize_slots,
            )
        else:
            atoms = [x]
        if vardb is None:
            # When called by repoman, we can simply return the first choice
            # because dep_eval() handles preference selection.
            return atoms

        all_available = True
        all_use_satisfied = True
        all_use_unmasked = True
        conflict_downgrade = False
        installed_downgrade = False
        slot_atoms = collections.defaultdict(list)
        slot_map = {}
        cp_map = {}
        for atom in atoms:
            if atom.blocker:
                continue

            # It's not a downgrade if parent is replacing child.
            replacing = (
                parent
                and graph_interface
                and graph_interface.will_replace_child(parent, myroot, atom)
            )
            # Ignore USE dependencies here since we don't want USE
            # settings to adversely affect || preference evaluation.
            avail_pkg = mydbapi_match_pkgs(atom.without_use)
            if not avail_pkg and replacing:
                avail_pkg = [replacing]
            if avail_pkg:
                avail_pkg = avail_pkg[-1]  # highest (ascending order)
                avail_slot = Atom(f"{atom.cp}:{avail_pkg.slot}")
            if not avail_pkg:
                all_available = False
                all_use_satisfied = False
                break

            if not replacing and graph_db is not None and downgrade_probe is not None:
                slot_matches = graph_db.match_pkgs(avail_slot)
                if (
                    len(slot_matches) > 1
                    and avail_pkg < slot_matches[-1]
                    and not downgrade_probe(avail_pkg)
                ):
                    # If a downgrade is not desirable, then avoid a
                    # choice that pulls in a lower version involved
                    # in a slot conflict (bug #531656).
                    conflict_downgrade = True

            if atom.use:
                avail_pkg_use = mydbapi_match_pkgs(atom)
                if not avail_pkg_use:
                    all_use_satisfied = False

                    if pkg_use_enabled is not None:
                        # Check which USE flags cause the match to fail,
                        # so we can prioritize choices that do not
                        # require changes to use.mask or use.force
                        # (see bug #515584).
                        violated_atom = atom.violated_conditionals(
                            pkg_use_enabled(avail_pkg), avail_pkg.iuse.is_valid_flag
                        )

                        # Note that violated_atom.use can be None here,
                        # since evaluation can collapse conditional USE
                        # deps that cause the match to fail due to
                        # missing IUSE (match uses atom.unevaluated_atom
                        # to detect such missing IUSE).
                        if violated_atom.use is not None:
                            for flag in violated_atom.use.enabled:
                                if flag in avail_pkg.use.mask:
                                    all_use_unmasked = False
                                    break
                            else:
                                for flag in violated_atom.use.disabled:
                                    if (
                                        flag in avail_pkg.use.force
                                        and flag not in avail_pkg.use.mask
                                    ):
                                        all_use_unmasked = False
                                        break
                else:
                    # highest (ascending order)
                    avail_pkg_use = avail_pkg_use[-1]
                    if avail_pkg_use != avail_pkg:
                        avail_pkg = avail_pkg_use
                    avail_slot = Atom(f"{atom.cp}:{avail_pkg.slot}")

            if not replacing and downgrade_probe is not None and graph is not None:
                highest_in_slot = mydbapi_match_pkgs(avail_slot)
                highest_in_slot = highest_in_slot[-1] if highest_in_slot else None
                if (
                    avail_pkg
                    and highest_in_slot
                    and avail_pkg < highest_in_slot
                    and not downgrade_probe(avail_pkg)
                    and (highest_in_slot.installed or highest_in_slot in graph)
                ):
                    installed_downgrade = True

            slot_map[avail_slot] = avail_pkg
            slot_atoms[avail_slot].append(atom)
            highest_cpv = cp_map.get(avail_pkg.cp)
            all_match_current = None
            all_match_previous = None
            if highest_cpv is not None and highest_cpv.slot == avail_pkg.slot:
                # If possible, make the package selection internally
                # consistent by choosing a package that satisfies all
                # atoms which match a package in the same slot. Later on,
                # the package version chosen here is used in the
                # has_upgrade/has_downgrade logic to prefer choices with
                # upgrades, and a package choice that is not internally
                # consistent will lead the has_upgrade/has_downgrade logic
                # to produce invalid results (see bug 600346).
                all_match_current = all(
                    a.match(avail_pkg) for a in slot_atoms[avail_slot]
                )
                all_match_previous = all(
                    a.match(highest_cpv) for a in slot_atoms[avail_slot]
                )
                if all_match_previous and not all_match_current:
                    continue

            current_higher = (
                highest_cpv is None
                or vercmp(avail_pkg.version, highest_cpv.version) > 0
            )

            if current_higher or (all_match_current and not all_match_previous):
                cp_map[avail_pkg.cp] = avail_pkg

        want_update = False
        if graph_interface is None or graph_interface.removal_action:
            new_slot_count = len(slot_map)
        else:
            new_slot_count = 0
            for slot_atom, avail_pkg in slot_map.items():
                if parent is not None and graph_interface.want_update_pkg(
                    parent, avail_pkg
                ):
                    want_update = True
                if not slot_atom.cp.startswith("virtual/") and not graph_db.match_pkgs(
                    slot_atom
                ):
                    new_slot_count += 1

        this_choice = _dep_choice(
            atoms=atoms,
            slot_map=slot_map,
            cp_map=cp_map,
            all_available=all_available,
            all_installed_slots=False,
            new_slot_count=new_slot_count,
            all_in_graph=False,
            want_update=want_update,
        )
        if all_available:
            # The "all installed" criterion is not version or slot specific.
            # If any version of a package is already in the graph then we
            # assume that it is preferred over other possible packages choices.
            all_installed = True
            for atom in {Atom(atom.cp) for atom in atoms if not atom.blocker}:
                # New-style virtuals have zero cost to install.
                if not vardb.match(atom) and not atom.startswith("virtual/"):
                    all_installed = False
                    break
            all_installed_slots = False
            if all_installed:
                all_installed_slots = True
                for slot_atom in slot_map:
                    # New-style virtuals have zero cost to install.
                    if not vardb.match(slot_atom) and not slot_atom.startswith(
                        "virtual/"
                    ):
                        all_installed_slots = False
                        break
            this_choice.all_installed_slots = all_installed_slots
            if graph_db is None:
                if all_use_satisfied:
                    if all_installed:
                        if all_installed_slots:
                            preferred_installed.append(this_choice)
                        else:
                            preferred_any_slot.append(this_choice)
                    else:
                        preferred_non_installed.append(this_choice)
                else:
                    if not all_use_unmasked:
                        other.append(this_choice)
                    elif all_installed_slots:
                        unsat_use_installed.append(this_choice)
                    else:
                        unsat_use_non_installed.append(this_choice)
            elif conflict_downgrade or installed_downgrade:
                other.append(this_choice)
            else:
                all_in_graph = True
                for atom in atoms:
                    # New-style virtuals have zero cost to install.
                    if atom.blocker or atom.cp.startswith("virtual/"):
                        continue
                    # We check if the matched package has actually been
                    # added to the digraph, in order to distinguish between
                    # those packages and installed packages that may need
                    # to be uninstalled in order to resolve blockers.
                    if not any(pkg in graph for pkg in graph_db.match_pkgs(atom)):
                        all_in_graph = False
                        break
                this_choice.all_in_graph = all_in_graph

                circular_atom = None
                if parent and parent.onlydeps:
                    # Check if the atom would result in a direct circular
                    # dependency and avoid that for --onlydeps arguments
                    # since it can defeat the purpose of --onlydeps.
                    # This check should only be used for --onlydeps
                    # arguments, since it can interfere with circular
                    # dependency backtracking choices, causing the test
                    # case for bug 756961 to fail.
                    cpv_slot_list = [parent]
                    for atom in atoms:
                        if atom.blocker:
                            continue
                        if vardb.match(atom):
                            # If the atom is satisfied by an installed
                            # version then it's not a circular dep.
                            continue
                        if atom.cp != parent.cp:
                            continue
                        if match_from_list(atom, cpv_slot_list):
                            circular_atom = atom
                            break
                if circular_atom is None and circular_dependency is not None:
                    for circular_child in itertools.chain(
                        circular_dependency.get(parent, []),
                        circular_dependency.get(virt_parent, []),
                    ):
                        for atom in atoms:
                            if not atom.blocker and atom.match(circular_child):
                                circular_atom = atom
                                break
                        if circular_atom is not None:
                            break

                if circular_atom is not None:
                    other.append(this_choice)
                else:
                    if all_use_satisfied:
                        if all_in_graph:
                            preferred_in_graph.append(this_choice)
                        elif all_installed:
                            if all_installed_slots:
                                preferred_installed.append(this_choice)
                            else:
                                preferred_any_slot.append(this_choice)
                        else:
                            preferred_non_installed.append(this_choice)
                    else:
                        if not all_use_unmasked:
                            other.append(this_choice)
                        elif all_in_graph:
                            unsat_use_in_graph.append(this_choice)
                        elif all_installed_slots:
                            unsat_use_installed.append(this_choice)
                        else:
                            unsat_use_non_installed.append(this_choice)
        else:
            all_installed = True
            some_installed = False
            for atom in atoms:
                if not atom.blocker:
                    if vardb.match(atom):
                        some_installed = True
                    else:
                        all_installed = False

            if all_installed:
                this_choice.all_installed_slots = True
                other_installed.append(this_choice)
            elif some_installed:
                other_installed_some.append(this_choice)

            # Use Atom(atom.cp) for a somewhat "fuzzy" match, since
            # the whole atom may be too specific. For example, see
            # bug #522652, where using the whole atom leads to an
            # unsatisfiable choice.
            elif any(vardb.match(Atom(atom.cp)) for atom in atoms if not atom.blocker):
                other_installed_any_slot.append(this_choice)
            else:
                other.append(this_choice)

    # Prefer choices which contain upgrades to higher slots. This helps
    # for deps such as || ( foo:1 foo:2 ), where we want to prefer the
    # atom which matches the higher version rather than the atom furthest
    # to the left. Sorting is done separately for each of choice_bins, so
    # as not to interfere with the ordering of the bins. Because of the
    # bin separation, the main function of this code is to allow
    # --depclean to remove old slots (rather than to pull in new slots).
    for choices in choice_bins:
        if len(choices) < 2:
            continue

        if minimize_slots:
            # Prefer choices having fewer new slots. When used with DNF form,
            # this can eliminate unnecessary packages that depclean would
            # ultimately eliminate (see bug 632026). Only use this behavior
            # when deemed necessary by the caller, since this will discard the
            # order specified in the ebuild, and the preferences specified
            # there can serve as a crucial sources of guidance (see bug 645002).

            # NOTE: Under some conditions, new_slot_count value may have some
            # variance from one calculation to the next because it depends on
            # the order that packages are added to the graph. This variance can
            # contribute to outcomes that appear to be random. Meanwhile,
            # the order specified in the ebuild is without variance, so it
            # does not have this problem.
            choices.sort(key=operator.attrgetter("new_slot_count"))

        for choice_1 in choices[1:]:
            cps = set(choice_1.cp_map)
            for choice_2 in choices:
                if choice_1 is choice_2:
                    # choice_1 will not be promoted, so move on
                    break
                if (
                    choice_1.all_installed_slots
                    and not choice_2.all_installed_slots
                    and not choice_2.want_update
                ):
                    # promote choice_1 in front of choice_2
                    choices.remove(choice_1)
                    index_2 = choices.index(choice_2)
                    choices.insert(index_2, choice_1)
                    break

                intersecting_cps = cps.intersection(choice_2.cp_map)
                has_upgrade = False
                has_downgrade = False
                for cp in intersecting_cps:
                    version_1 = choice_1.cp_map[cp]
                    version_2 = choice_2.cp_map[cp]
                    difference = vercmp(version_1.version, version_2.version)
                    if difference != 0:
                        if difference > 0:
                            has_upgrade = True
                        else:
                            has_downgrade = True

                if (
                    # Prefer upgrades.
                    (has_upgrade and not has_downgrade)
                    # Prefer choices where all packages have been pulled into
                    # the graph, except for choices that eliminate upgrades.
                    or (
                        choice_1.all_in_graph
                        and not choice_2.all_in_graph
                        and not (has_downgrade and not has_upgrade)
                    )
                ):
                    # promote choice_1 in front of choice_2
                    choices.remove(choice_1)
                    index_2 = choices.index(choice_2)
                    choices.insert(index_2, choice_1)
                    break

    for allow_masked in (False, True):
        for choices in choice_bins:
            for choice in choices:
                if choice.all_available or allow_masked:
                    return choice.atoms

    assert False  # This point should not be reachable


def dep_check(
    depstring,
    mydbapi,
    mysettings,
    use="yes",
    mode=None,
    myuse=None,
    use_cache=1,
    use_binaries=0,
    myroot=None,
    trees=None,
):
    """
    Takes a depend string, parses it, and selects atoms.
    The myroot parameter is unused (use mysettings['EROOT'] instead).
    """
    myroot = mysettings["EROOT"]
    edebug = mysettings.get("PORTAGE_DEBUG", None) == "1"
    # check_config_instance(mysettings)
    if trees is None:
        trees = globals()["db"]
    if use == "yes":
        if myuse is None:
            # default behavior
            myusesplit = mysettings["PORTAGE_USE"].split()
        else:
            myusesplit = myuse
            # We've been given useflags to use.
            # print "USE FLAGS PASSED IN."
            # print myuse
            # if "bindist" in myusesplit:
            # 	print "BINDIST is set!"
            # else:
            # 	print "BINDIST NOT set."
    else:
        # we are being run by autouse(), don't consult USE vars yet.
        # WE ALSO CANNOT USE SETTINGS
        myusesplit = []

    mymasks = set()
    useforce = set()
    if use == "all":
        # This is only for repoman, in order to constrain the use_reduce
        # matchall behavior to account for profile use.mask/force. The
        # ARCH/archlist code here may be redundant, since the profile
        # really should be handling ARCH masking/forcing itself.
        arch = mysettings.get("ARCH")
        mymasks.update(mysettings.usemask)
        mymasks.update(mysettings.archlist())
        if arch:
            mymasks.discard(arch)
            useforce.add(arch)
        useforce.update(mysettings.useforce)
        useforce.difference_update(mymasks)

    # eapi code borrowed from _expand_new_virtuals()
    mytrees = trees[myroot]
    parent = mytrees.get("parent")
    virt_parent = mytrees.get("virt_parent")
    current_parent = None
    eapi = None
    if parent is not None:
        if virt_parent is not None:
            current_parent = virt_parent
        else:
            current_parent = parent

    if current_parent is not None:
        # Don't pass the eapi argument to use_reduce() for installed packages
        # since previous validation will have already marked them as invalid
        # when necessary and now we're more interested in evaluating
        # dependencies so that things like --depclean work as well as possible
        # in spite of partial invalidity.
        if not current_parent.installed:
            eapi = current_parent.eapi

    if isinstance(depstring, list):
        mysplit = depstring
    else:
        try:
            mysplit = use_reduce(
                depstring,
                uselist=myusesplit,
                masklist=mymasks,
                matchall=(use == "all"),
                excludeall=useforce,
                opconvert=True,
                token_class=Atom,
                eapi=eapi,
            )
        except InvalidDependString as e:
            return [0, f"{e}"]

    if mysplit == []:
        # dependencies were reduced to nothing
        return [1, []]

    # Recursively expand new-style virtuals so as to
    # collapse one or more levels of indirection.
    try:
        mysplit = _expand_new_virtuals(
            mysplit,
            edebug,
            mydbapi,
            mysettings,
            use=use,
            mode=mode,
            myuse=myuse,
            use_force=useforce,
            use_mask=mymasks,
            use_cache=use_cache,
            use_binaries=use_binaries,
            myroot=myroot,
            trees=trees,
        )
    except ParseError as e:
        return [0, f"{e}"]

    dnf = False
    if mysettings.local_config:  # if not repoman
        orig_split = mysplit
        mysplit = _overlap_dnf(mysplit)
        dnf = mysplit is not orig_split

    mysplit2 = dep_wordreduce(mysplit, mysettings, mydbapi, mode, use_cache=use_cache)
    if mysplit2 is None:
        return [0, _("Invalid token")]

    writemsg("\n\n\n", 1)
    writemsg(f"mysplit:  {mysplit}\n", 1)
    writemsg(f"mysplit2: {mysplit2}\n", 1)

    selected_atoms = dep_zapdeps(
        mysplit,
        mysplit2,
        myroot,
        use_binaries=use_binaries,
        trees=trees,
        minimize_slots=dnf,
    )

    return [1, selected_atoms]


def _overlap_dnf(dep_struct):
    """
    Combine overlapping || groups using disjunctive normal form (DNF), in
    order to minimize the number of packages chosen to satisfy cases like
    "|| ( foo bar ) || ( bar baz )" as in bug #632026. Non-overlapping
    groups are excluded from the conversion, since DNF leads to exponential
    explosion of the formula.

    When dep_struct does not contain any overlapping groups, no DNF
    conversion will be performed, and dep_struct will be returned as-is.
    Callers can detect this case by checking if the returned object has
    the same identity as dep_struct. If the identity is different, then
    DNF conversion was performed.
    """
    if not _contains_disjunction(dep_struct):
        return dep_struct

    # map atom.cp to disjunctions
    cp_map = collections.defaultdict(list)
    # graph atom.cp, with edges connecting atoms in the same disjunction
    overlap_graph = digraph()
    # map id(disjunction) to index in dep_struct, for deterministic output
    order_map = {}
    order_key = lambda x: order_map[id(x)]
    result = []
    for i, x in enumerate(dep_struct):
        if isinstance(x, list):
            assert (
                x and x[0] == "||"
            ), f"Normalization error, nested conjunction found in {dep_struct}"
            order_map[id(x)] = i
            prev_cp = None
            for atom in _iter_flatten(x):
                if isinstance(atom, Atom) and not atom.blocker:
                    cp_map[atom.cp].append(x)
                    overlap_graph.add(atom.cp, parent=prev_cp)
                    prev_cp = atom.cp
            if prev_cp is None:  # only contains blockers
                result.append(x)
        else:
            result.append(x)

    # group together disjunctions having atom.cp overlap
    traversed = set()
    overlap = False
    for cp in overlap_graph:
        if cp in traversed:
            continue
        disjunctions = {}
        stack = [cp]
        while stack:
            cp = stack.pop()
            traversed.add(cp)
            for x in cp_map[cp]:
                disjunctions[id(x)] = x
            for other_cp in itertools.chain(
                overlap_graph.child_nodes(cp), overlap_graph.parent_nodes(cp)
            ):
                if other_cp not in traversed:
                    stack.append(other_cp)

        if len(disjunctions) > 1:
            overlap = True
            # convert overlapping disjunctions to DNF
            result.extend(_dnf_convert(sorted(disjunctions.values(), key=order_key)))
        else:
            # pass through non-overlapping disjunctions
            result.append(disjunctions.popitem()[1])

    return result if overlap else dep_struct


def _iter_flatten(dep_struct):
    """
    Yield nested elements of dep_struct.
    """
    for x in dep_struct:
        if isinstance(x, list):
            for x in _iter_flatten(x):
                yield x
        else:
            yield x


def dep_wordreduce(mydeplist, mysettings, mydbapi, mode, use_cache=1):
    "Reduces the deplist to ones and zeros"
    deplist = mydeplist[:]
    for mypos, token in enumerate(deplist):
        if isinstance(deplist[mypos], list):
            # recurse
            deplist[mypos] = dep_wordreduce(
                deplist[mypos], mysettings, mydbapi, mode, use_cache=use_cache
            )
        elif deplist[mypos] == "||":
            pass
        elif token[:1] == "!":
            deplist[mypos] = False
        else:
            mykey = deplist[mypos].cp
            if (
                mysettings
                and mykey in mysettings.pprovideddict
                and match_from_list(deplist[mypos], mysettings.pprovideddict[mykey])
            ):
                deplist[mypos] = True
            elif mydbapi is None:
                # Assume nothing is satisfied.  This forces dep_zapdeps to
                # return all of deps the deps that have been selected
                # (excluding those satisfied by package.provided).
                deplist[mypos] = False
            else:
                if mode:
                    x = mydbapi.xmatch(mode, deplist[mypos])
                    if mode.startswith("minimum-"):
                        mydep = []
                        if x:
                            mydep.append(x)
                    else:
                        mydep = x
                else:
                    mydep = mydbapi.match(deplist[mypos], use_cache=use_cache)
                if mydep is not None:
                    tmp = len(mydep) >= 1
                    if deplist[mypos][0] == "!":
                        tmp = False
                    deplist[mypos] = tmp
                else:
                    # encountered invalid string
                    return None
    return deplist



# Copyright 2019-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import io
import tempfile
import types

import portage
from portage import shutil, os
from portage.checksum import checksum_str
from portage.const import BASH_BINARY, MANIFEST2_HASH_DEFAULTS, PORTAGE_PYM_PATH
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.tests.util.test_socks5 import AsyncHTTPServer
from portage.util.configparser import ConfigParserError
from portage.util.futures import asyncio
from portage.util.futures.executor.fork import ForkExecutor
from portage.util._async.SchedulerInterface import SchedulerInterface
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.package.ebuild.config import config
from portage.package.ebuild.digestgen import digestgen
from portage.package.ebuild.fetch import (
    ContentHashLayout,
    DistfileName,
    _download_suffix,
    fetch,
    FilenameHashLayout,
    FlatLayout,
    MirrorLayoutConfig,
)
from portage._emirrordist.Config import Config as EmirrordistConfig
from _emerge.EbuildFetcher import EbuildFetcher
from _emerge.Package import Package


class EbuildFetchTestCase(TestCase):
    def testEbuildFetch(self):
        user_config = {
            "make.conf": ('GENTOO_MIRRORS="{scheme}://{host}:{port}"',),
        }

        distfiles = {
            "bar": b"bar\n",
            "foo": b"foo\n",
        }

        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "7",
                "SRC_URI": """{scheme}://{host}:{port}/distfiles/bar.txt -> bar
					{scheme}://{host}:{port}/distfiles/foo.txt -> foo""",
            },
        }

        loop = SchedulerInterface(global_event_loop())

        scheme = "http"
        host = "127.0.0.1"
        content = {}

        with AsyncHTTPServer(host, content, loop) as server:
            ebuilds_subst = {}
            for cpv, metadata in ebuilds.items():
                metadata = metadata.copy()
                metadata["SRC_URI"] = metadata["SRC_URI"].format(
                    scheme=scheme, host=host, port=server.server_port
                )
                ebuilds_subst[cpv] = metadata

            user_config_subst = user_config.copy()
            for configname, configdata in user_config.items():
                configdata_sub = []
                for line in configdata:
                    configdata_sub.append(
                        line.format(scheme=scheme, host=host, port=server.server_port)
                    )
                user_config_subst[configname] = tuple(configdata_sub)

            playground = ResolverPlayground(
                ebuilds=ebuilds_subst,
                distfiles=distfiles,
                user_config=user_config_subst,
            )
            ro_distdir = tempfile.mkdtemp()
            try:
                self._testEbuildFetch(
                    loop,
                    scheme,
                    host,
                    distfiles,
                    ebuilds,
                    content,
                    server,
                    playground,
                    ro_distdir,
                )
            finally:
                shutil.rmtree(ro_distdir)
                playground.cleanup()

    def _testEbuildFetch(
        self,
        loop,
        scheme,
        host,
        orig_distfiles,
        ebuilds,
        content,
        server,
        playground,
        ro_distdir,
    ):
        mirror_layouts = (
            (
                "[structure]",
                "0=filename-hash BLAKE2B 8",
                "1=flat",
            ),
            (
                "[structure]",
                "1=filename-hash BLAKE2B 8",
                "0=flat",
            ),
            (
                "[structure]",
                "0=content-hash SHA512 8:8:8",
                "1=flat",
            ),
        )

        fetchcommand = portage.util.shlex_split(playground.settings["FETCHCOMMAND"])
        fetch_bin = portage.process.find_binary(fetchcommand[0])
        if fetch_bin is None:
            self.skipTest(
                f"FETCHCOMMAND not found: {playground.settings['FETCHCOMMAND']}"
            )
        eubin = os.path.join(playground.eprefix, "usr", "bin")
        os.symlink(fetch_bin, os.path.join(eubin, os.path.basename(fetch_bin)))
        resumecommand = portage.util.shlex_split(playground.settings["RESUMECOMMAND"])
        resume_bin = portage.process.find_binary(resumecommand[0])
        if resume_bin is None:
            self.skipTest(
                f"RESUMECOMMAND not found: {playground.settings['RESUMECOMMAND']}"
            )
        if resume_bin != fetch_bin:
            os.symlink(resume_bin, os.path.join(eubin, os.path.basename(resume_bin)))
        root_config = playground.trees[playground.eroot]["root_config"]
        portdb = root_config.trees["porttree"].dbapi

        def run_async(func, *args, **kwargs):
            with ForkExecutor(loop=loop) as executor:
                return loop.run_until_complete(
                    loop.run_in_executor(
                        executor, functools.partial(func, *args, **kwargs)
                    )
                )

        for layout_lines in mirror_layouts:
            settings = config(clone=playground.settings)
            layout_data = "".join(f"{line}\n" for line in layout_lines)
            mirror_conf = MirrorLayoutConfig()
            mirror_conf.read_from_file(io.StringIO(layout_data))
            layouts = mirror_conf.get_all_layouts()
            content["/distfiles/layout.conf"] = layout_data.encode("utf8")
            distfiles = {}
            for k, v in orig_distfiles.items():
                filename = DistfileName(
                    k,
                    digests={
                        algo: checksum_str(v, hashname=algo)
                        for algo in MANIFEST2_HASH_DEFAULTS
                    },
                )
                distfiles[filename] = v

                # mirror path
                for layout in layouts:
                    content["/distfiles/" + layout.get_path(filename)] = v
                # upstream path
                content[f"/distfiles/{k}.txt"] = v

            shutil.rmtree(settings["DISTDIR"])
            os.makedirs(settings["DISTDIR"])
            with open(os.path.join(settings["DISTDIR"], "layout.conf"), "w") as f:
                f.write(layout_data)

            if any(isinstance(layout, ContentHashLayout) for layout in layouts):
                content_db = os.path.join(
                    playground.eprefix, "var/db/emirrordist/content.db"
                )
                os.makedirs(os.path.dirname(content_db), exist_ok=True)
                try:
                    os.unlink(content_db)
                except OSError:
                    pass
            else:
                content_db = None

            # Demonstrate that fetch preserves a stale file in DISTDIR when no digests are given.
            foo_uri = {
                "foo": (f"{scheme}://{host}:{server.server_port}/distfiles/foo",)
            }
            foo_path = os.path.join(settings["DISTDIR"], "foo")
            foo_stale_content = b"stale content\n"
            with open(foo_path, "wb") as f:
                f.write(b"stale content\n")

            self.assertTrue(
                bool(run_async(fetch, foo_uri, settings, try_mirrors=False))
            )

            with open(foo_path, "rb") as f:
                self.assertEqual(f.read(), foo_stale_content)
            with open(foo_path, "rb") as f:
                self.assertNotEqual(f.read(), distfiles["foo"])

            # Use force=True to update the stale file.
            self.assertTrue(
                bool(run_async(fetch, foo_uri, settings, try_mirrors=False, force=True))
            )

            with open(foo_path, "rb") as f:
                self.assertEqual(f.read(), distfiles["foo"])

            # Test force=True with FEATURES=skiprocheck, using read-only DISTDIR.
            # FETCHCOMMAND is set to temporarily chmod +w DISTDIR. Note that
            # FETCHCOMMAND must perform atomic rename itself due to read-only
            # DISTDIR.
            with open(foo_path, "wb") as f:
                f.write(b"stale content\n")
            orig_fetchcommand = settings["FETCHCOMMAND"]
            orig_distdir_mode = os.stat(settings["DISTDIR"]).st_mode
            temp_fetchcommand = os.path.join(eubin, "fetchcommand")
            with open(temp_fetchcommand, "w") as f:
                f.write(
                    """
					set -e
					URI=$1
					DISTDIR=$2
					FILE=$3
					trap 'chmod a-w "${DISTDIR}"' EXIT
					chmod ug+w "${DISTDIR}"
					%s
					mv -f "${DISTDIR}/${FILE}.__download__" "${DISTDIR}/${FILE}"
				"""
                    % orig_fetchcommand.replace("${FILE}", "${FILE}.__download__")
                )
            settings[
                "FETCHCOMMAND"
            ] = '"{}" "{}" "${{URI}}" "${{DISTDIR}}" "${{FILE}}"'.format(
                BASH_BINARY,
                temp_fetchcommand,
            )
            settings.features.add("skiprocheck")
            settings.features.remove("distlocks")
            os.chmod(settings["DISTDIR"], 0o555)
            try:
                self.assertTrue(
                    bool(
                        run_async(
                            fetch, foo_uri, settings, try_mirrors=False, force=True
                        )
                    )
                )
            finally:
                settings["FETCHCOMMAND"] = orig_fetchcommand
                os.chmod(settings["DISTDIR"], orig_distdir_mode)
                settings.features.remove("skiprocheck")
                settings.features.add("distlocks")
                os.unlink(temp_fetchcommand)

            with open(foo_path, "rb") as f:
                self.assertEqual(f.read(), distfiles["foo"])

            # Test emirrordist invocation.
            emirrordist_cmd = (
                portage._python_interpreter,
                "-b",
                "-Wd",
                os.path.join(str(self.bindir), "emirrordist"),
                "--distfiles",
                settings["DISTDIR"],
                "--config-root",
                settings["EPREFIX"],
                "--delete",
                "--repositories-configuration",
                settings.repositories.config_string(),
                "--repo",
                "test_repo",
                "--mirror",
            )

            if content_db is not None:
                emirrordist_cmd = emirrordist_cmd + (
                    "--content-db",
                    content_db,
                )

            env = settings.environ()
            env["PYTHONPATH"] = ":".join(
                filter(
                    None,
                    [PORTAGE_PYM_PATH] + os.environ.get("PYTHONPATH", "").split(":"),
                )
            )

            for k in distfiles:
                try:
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                except OSError:
                    pass

            proc = loop.run_until_complete(
                asyncio.create_subprocess_exec(*emirrordist_cmd, env=env)
            )
            self.assertEqual(loop.run_until_complete(proc.wait()), 0)

            for k in distfiles:
                with open(
                    os.path.join(settings["DISTDIR"], layouts[0].get_path(k)), "rb"
                ) as f:
                    self.assertEqual(f.read(), distfiles[k])

            if content_db is not None:
                loop.run_until_complete(
                    self._test_content_db(
                        emirrordist_cmd,
                        env,
                        layouts,
                        content_db,
                        distfiles,
                        settings,
                        portdb,
                    )
                )

            config_pool = config_pool_cls(settings)

            def async_fetch(pkg, ebuild_path):
                fetcher = EbuildFetcher(
                    config_pool=config_pool,
                    ebuild_path=ebuild_path,
                    fetchonly=False,
                    fetchall=True,
                    pkg=pkg,
                    scheduler=loop,
                )
                fetcher.start()
                return fetcher.async_wait()

            for cpv in ebuilds:
                metadata = dict(
                    zip(
                        Package.metadata_keys,
                        portdb.aux_get(cpv, Package.metadata_keys),
                    )
                )

                pkg = Package(
                    built=False,
                    cpv=cpv,
                    installed=False,
                    metadata=metadata,
                    root_config=root_config,
                    type_name="ebuild",
                )

                settings.setcpv(pkg)
                ebuild_path = portdb.findname(pkg.cpv)
                portage.doebuild_environment(
                    ebuild_path, "fetch", settings=settings, db=portdb
                )

                # Test good files in DISTDIR
                for k in settings["AA"].split():
                    os.stat(os.path.join(settings["DISTDIR"], k))
                self.assertEqual(
                    loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test digestgen with fetch
                os.unlink(os.path.join(os.path.dirname(ebuild_path), "Manifest"))
                for k in settings["AA"].split():
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                with ForkExecutor(loop=loop) as executor:
                    self.assertTrue(
                        bool(
                            loop.run_until_complete(
                                loop.run_in_executor(
                                    executor,
                                    functools.partial(
                                        digestgen, mysettings=settings, myportdb=portdb
                                    ),
                                )
                            )
                        )
                    )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test missing files in DISTDIR
                for k in settings["AA"].split():
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                self.assertEqual(
                    loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test empty files in DISTDIR
                for k in settings["AA"].split():
                    file_path = os.path.join(settings["DISTDIR"], k)
                    with open(file_path, "wb") as f:
                        pass
                    self.assertEqual(os.stat(file_path).st_size, 0)
                self.assertEqual(
                    loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test non-empty files containing null bytes in DISTDIR
                for k in settings["AA"].split():
                    file_path = os.path.join(settings["DISTDIR"], k)
                    with open(file_path, "wb") as f:
                        f.write(len(distfiles[k]) * b"\0")
                    self.assertEqual(os.stat(file_path).st_size, len(distfiles[k]))
                self.assertEqual(
                    loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                )
                for k in settings["AA"].split():
                    with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                        self.assertEqual(f.read(), distfiles[k])

                # Test PORTAGE_RO_DISTDIRS
                settings["PORTAGE_RO_DISTDIRS"] = f'"{ro_distdir}"'
                orig_fetchcommand = settings["FETCHCOMMAND"]
                orig_resumecommand = settings["RESUMECOMMAND"]
                try:
                    settings["FETCHCOMMAND"] = settings["RESUMECOMMAND"] = ""
                    for k in settings["AA"].split():
                        file_path = os.path.join(settings["DISTDIR"], k)
                        os.rename(file_path, os.path.join(ro_distdir, k))
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        file_path = os.path.join(settings["DISTDIR"], k)
                        self.assertTrue(os.path.islink(file_path))
                        with open(file_path, "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                        os.unlink(file_path)
                finally:
                    settings.pop("PORTAGE_RO_DISTDIRS")
                    settings["FETCHCOMMAND"] = orig_fetchcommand
                    settings["RESUMECOMMAND"] = orig_resumecommand

                # Test local filesystem in GENTOO_MIRRORS
                orig_mirrors = settings["GENTOO_MIRRORS"]
                orig_fetchcommand = settings["FETCHCOMMAND"]
                try:
                    settings["GENTOO_MIRRORS"] = ro_distdir
                    settings["FETCHCOMMAND"] = settings["RESUMECOMMAND"] = ""
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                finally:
                    settings["GENTOO_MIRRORS"] = orig_mirrors
                    settings["FETCHCOMMAND"] = orig_fetchcommand
                    settings["RESUMECOMMAND"] = orig_resumecommand

                # Test readonly DISTDIR
                orig_distdir_mode = os.stat(settings["DISTDIR"]).st_mode
                try:
                    os.chmod(settings["DISTDIR"], 0o555)
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                finally:
                    os.chmod(settings["DISTDIR"], orig_distdir_mode)

                # Test parallel-fetch mode
                settings["PORTAGE_PARALLEL_FETCHONLY"] = "1"
                try:
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                    for k in settings["AA"].split():
                        os.unlink(os.path.join(settings["DISTDIR"], k))
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                finally:
                    settings.pop("PORTAGE_PARALLEL_FETCHONLY")

                # Test RESUMECOMMAND
                orig_resume_min_size = settings["PORTAGE_FETCH_RESUME_MIN_SIZE"]
                try:
                    settings["PORTAGE_FETCH_RESUME_MIN_SIZE"] = "2"
                    for k in settings["AA"].split():
                        file_path = os.path.join(settings["DISTDIR"], k)
                        os.unlink(file_path)
                        with open(file_path + _download_suffix, "wb") as f:
                            f.write(distfiles[k][:2])
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                    for k in settings["AA"].split():
                        with open(os.path.join(settings["DISTDIR"], k), "rb") as f:
                            self.assertEqual(f.read(), distfiles[k])
                finally:
                    settings["PORTAGE_FETCH_RESUME_MIN_SIZE"] = orig_resume_min_size

                # Test readonly DISTDIR + skiprocheck, with FETCHCOMMAND set to temporarily chmod DISTDIR
                orig_fetchcommand = settings["FETCHCOMMAND"]
                orig_distdir_mode = os.stat(settings["DISTDIR"]).st_mode
                for k in settings["AA"].split():
                    os.unlink(os.path.join(settings["DISTDIR"], k))
                try:
                    os.chmod(settings["DISTDIR"], 0o555)
                    settings["FETCHCOMMAND"] = (
                        '"%s" -c "chmod ug+w \\"${DISTDIR}\\"; %s; status=\\$?; chmod a-w \\"${DISTDIR}\\"; exit \\$status"'
                        % (BASH_BINARY, orig_fetchcommand.replace('"', '\\"'))
                    )
                    settings.features.add("skiprocheck")
                    settings.features.remove("distlocks")
                    self.assertEqual(
                        loop.run_until_complete(async_fetch(pkg, ebuild_path)), 0
                    )
                finally:
                    settings["FETCHCOMMAND"] = orig_fetchcommand
                    os.chmod(settings["DISTDIR"], orig_distdir_mode)
                    settings.features.remove("skiprocheck")
                    settings.features.add("distlocks")

    async def _test_content_db(
        self, emirrordist_cmd, env, layouts, content_db, distfiles, settings, portdb
    ):
        # Simulate distfile digest change for ContentDB.
        emdisopts = types.SimpleNamespace(
            content_db=content_db, distfiles=settings["DISTDIR"]
        )
        with EmirrordistConfig(
            emdisopts, portdb, asyncio.get_event_loop()
        ) as emdisconf:
            # Copy revisions from bar to foo.
            for revision_key in emdisconf.content_db["filename:bar"]:
                emdisconf.content_db.add(
                    DistfileName("foo", digests=dict(revision_key))
                )

            # Copy revisions from foo to bar.
            for revision_key in emdisconf.content_db["filename:foo"]:
                emdisconf.content_db.add(
                    DistfileName("bar", digests=dict(revision_key))
                )

            content_db_state = dict(emdisconf.content_db.items())
            self.assertEqual(content_db_state, dict(emdisconf.content_db.items()))
            self.assertEqual(
                {
                    k[len("filename:") :]
                    for k in content_db_state
                    if k.startswith("filename:")
                },
                {"bar", "foo"},
            )
            self.assertEqual(
                content_db_state["filename:foo"], content_db_state["filename:bar"]
            )
            self.assertEqual(len(content_db_state["filename:foo"]), 2)

        for k in distfiles:
            try:
                os.unlink(os.path.join(settings["DISTDIR"], k))
            except OSError:
                pass

        proc = await asyncio.create_subprocess_exec(*emirrordist_cmd, env=env)
        self.assertEqual(await proc.wait(), 0)

        for k in distfiles:
            with open(
                os.path.join(settings["DISTDIR"], layouts[0].get_path(k)), "rb"
            ) as f:
                self.assertEqual(f.read(), distfiles[k])

        with EmirrordistConfig(
            emdisopts, portdb, asyncio.get_event_loop()
        ) as emdisconf:
            self.assertEqual(content_db_state, dict(emdisconf.content_db.items()))

            # Verify that remove works as expected
            filename = [filename for filename in distfiles if filename == "foo"][0]
            self.assertTrue(bool(filename.digests))
            emdisconf.content_db.remove(filename)
            # foo should still have a content revision corresponding to bar's content.
            self.assertEqual(
                {
                    k[len("filename:") :]
                    for k in emdisconf.content_db
                    if k.startswith("filename:")
                },
                {"bar", "foo"},
            )
            self.assertEqual(len(emdisconf.content_db["filename:foo"]), 1)
            self.assertEqual(
                len(
                    [
                        revision_key
                        for revision_key in emdisconf.content_db["filename:foo"]
                        if not filename.digests_equal(
                            DistfileName(
                                "foo",
                                digests=dict(revision_key),
                            )
                        )
                    ]
                ),
                1,
            )
            # bar should still have a content revision corresponding to foo's content.
            self.assertEqual(len(emdisconf.content_db["filename:bar"]), 2)
            self.assertEqual(
                len(
                    [
                        revision_key
                        for revision_key in emdisconf.content_db["filename:bar"]
                        if filename.digests_equal(
                            DistfileName(
                                "bar",
                                digests=dict(revision_key),
                            )
                        )
                    ]
                ),
                1,
            )
            # remove the foo which refers to bar's content
            bar = [filename for filename in distfiles if filename == "bar"][0]
            foo_remaining = DistfileName("foo", digests=bar.digests)
            emdisconf.content_db.remove(foo_remaining)
            self.assertEqual(
                [
                    k[len("filename:") :]
                    for k in emdisconf.content_db
                    if k.startswith("filename:")
                ],
                ["bar"],
            )
            self.assertRaises(
                KeyError, emdisconf.content_db.__getitem__, "filename:foo"
            )
            # bar should still have a content revision corresponding to foo's content.
            self.assertEqual(len(emdisconf.content_db["filename:bar"]), 2)

    def test_flat_layout(self):
        self.assertTrue(FlatLayout.verify_args(("flat",)))
        self.assertFalse(FlatLayout.verify_args(("flat", "extraneous-arg")))
        self.assertEqual(FlatLayout().get_path("foo-1.tar.gz"), "foo-1.tar.gz")

    def test_filename_hash_layout(self):
        self.assertFalse(FilenameHashLayout.verify_args(("filename-hash",)))
        self.assertTrue(FilenameHashLayout.verify_args(("filename-hash", "SHA1", "8")))
        self.assertFalse(
            FilenameHashLayout.verify_args(("filename-hash", "INVALID-HASH", "8"))
        )
        self.assertTrue(
            FilenameHashLayout.verify_args(("filename-hash", "SHA1", "4:8:12"))
        )
        self.assertFalse(FilenameHashLayout.verify_args(("filename-hash", "SHA1", "3")))
        self.assertFalse(
            FilenameHashLayout.verify_args(("filename-hash", "SHA1", "junk"))
        )
        self.assertFalse(
            FilenameHashLayout.verify_args(("filename-hash", "SHA1", "4:8:junk"))
        )

        self.assertEqual(
            FilenameHashLayout("SHA1", "4").get_path("foo-1.tar.gz"), "1/foo-1.tar.gz"
        )
        self.assertEqual(
            FilenameHashLayout("SHA1", "8").get_path("foo-1.tar.gz"), "19/foo-1.tar.gz"
        )
        self.assertEqual(
            FilenameHashLayout("SHA1", "8:16").get_path("foo-1.tar.gz"),
            "19/c3b6/foo-1.tar.gz",
        )
        self.assertEqual(
            FilenameHashLayout("SHA1", "8:16:24").get_path("foo-1.tar.gz"),
            "19/c3b6/37a94b/foo-1.tar.gz",
        )

    def test_content_hash_layout(self):
        self.assertFalse(ContentHashLayout.verify_args(("content-hash",)))
        self.assertTrue(ContentHashLayout.verify_args(("content-hash", "SHA1", "8")))
        self.assertFalse(
            ContentHashLayout.verify_args(("content-hash", "INVALID-HASH", "8"))
        )
        self.assertTrue(
            ContentHashLayout.verify_args(("content-hash", "SHA1", "4:8:12"))
        )
        self.assertFalse(ContentHashLayout.verify_args(("content-hash", "SHA1", "3")))
        self.assertFalse(
            ContentHashLayout.verify_args(("content-hash", "SHA1", "junk"))
        )
        self.assertFalse(
            ContentHashLayout.verify_args(("content-hash", "SHA1", "4:8:junk"))
        )

        filename = DistfileName(
            "foo-1.tar.gz",
            digests={
                algo: checksum_str(b"", hashname=algo)
                for algo in MANIFEST2_HASH_DEFAULTS
            },
        )

        # Raise KeyError for a hash algorithm SHA1 which is not in MANIFEST2_HASH_DEFAULTS.
        self.assertRaises(KeyError, ContentHashLayout("SHA1", "4").get_path, filename)

        # Raise AttributeError for a plain string argument.
        self.assertRaises(
            AttributeError, ContentHashLayout("SHA512", "4").get_path, str(filename)
        )

        self.assertEqual(
            ContentHashLayout("SHA512", "4").get_path(filename),
            "c/cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )
        self.assertEqual(
            ContentHashLayout("SHA512", "8").get_path(filename),
            "cf/cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )
        self.assertEqual(
            ContentHashLayout("SHA512", "8:16").get_path(filename),
            "cf/83e1/cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )
        self.assertEqual(
            ContentHashLayout("SHA512", "8:16:24").get_path(filename),
            "cf/83e1/357eef/cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )

    def test_mirror_layout_config(self):
        mlc = MirrorLayoutConfig()
        self.assertEqual(mlc.serialize(), ())
        self.assertIsInstance(mlc.get_best_supported_layout(), FlatLayout)

        conf = """
[structure]
0=flat
"""
        mlc.read_from_file(io.StringIO(conf))
        self.assertEqual(mlc.serialize(), (("flat",),))
        self.assertIsInstance(mlc.get_best_supported_layout(), FlatLayout)
        self.assertEqual(
            mlc.get_best_supported_layout().get_path("foo-1.tar.gz"), "foo-1.tar.gz"
        )

        conf = """
[structure]
0=filename-hash SHA1 8:16
1=flat
"""
        mlc.read_from_file(io.StringIO(conf))
        self.assertEqual(
            mlc.serialize(), (("filename-hash", "SHA1", "8:16"), ("flat",))
        )
        self.assertIsInstance(mlc.get_best_supported_layout(), FilenameHashLayout)
        self.assertEqual(
            mlc.get_best_supported_layout().get_path("foo-1.tar.gz"),
            "19/c3b6/foo-1.tar.gz",
        )
        serialized = mlc.serialize()

        # test fallback
        conf = """
[structure]
0=filename-hash INVALID-HASH 8:16
1=filename-hash SHA1 32
2=flat
"""
        mlc.read_from_file(io.StringIO(conf))
        self.assertEqual(
            mlc.serialize(),
            (
                ("filename-hash", "INVALID-HASH", "8:16"),
                ("filename-hash", "SHA1", "32"),
                ("flat",),
            ),
        )
        self.assertIsInstance(mlc.get_best_supported_layout(), FilenameHashLayout)
        self.assertEqual(
            mlc.get_best_supported_layout().get_path("foo-1.tar.gz"),
            "19c3b637/foo-1.tar.gz",
        )

        # test deserialization
        mlc.deserialize(serialized)
        self.assertEqual(
            mlc.serialize(), (("filename-hash", "SHA1", "8:16"), ("flat",))
        )
        self.assertIsInstance(mlc.get_best_supported_layout(), FilenameHashLayout)
        self.assertEqual(
            mlc.get_best_supported_layout().get_path("foo-1.tar.gz"),
            "19/c3b6/foo-1.tar.gz",
        )

        # test erraneous input
        conf = """
[#(*DA*&*F
[structure]
0=filename-hash SHA1 32
"""
        self.assertRaises(ConfigParserError, mlc.read_from_file, io.StringIO(conf))

    def test_filename_hash_layout_get_filenames(self):
        filename = DistfileName(
            "foo-1.tar.gz",
            digests={
                algo: checksum_str(b"", hashname=algo)
                for algo in MANIFEST2_HASH_DEFAULTS
            },
        )
        layouts = (
            FlatLayout(),
            FilenameHashLayout("SHA1", "4"),
            FilenameHashLayout("SHA1", "8"),
            FilenameHashLayout("SHA1", "8:16"),
            FilenameHashLayout("SHA1", "8:16:24"),
            ContentHashLayout("SHA512", "8:8:8"),
        )

        for layout in layouts:
            distdir = tempfile.mkdtemp()
            try:
                path = os.path.join(distdir, layout.get_path(filename))
                try:
                    os.makedirs(os.path.dirname(path))
                except OSError:
                    pass

                with open(path, "wb") as f:
                    pass

                file_list = list(layout.get_filenames(distdir))
                self.assertTrue(len(file_list) > 0)
                for filename_result in file_list:
                    if isinstance(filename_result, DistfileName):
                        self.assertTrue(filename_result.digests_equal(filename))
                    else:
                        self.assertEqual(filename_result, str(filename))
            finally:
                shutil.rmtree(distdir)


# Tests only work with one ebuild at a time, so the config
# pool only needs a single config instance.
class config_pool_cls:
    def __init__(self, settings):
        self._settings = settings

    def allocate(self):
        return self._settings

    def deallocate(self, settings):
        pass
